"""Observation."""
import json

import numpy as np


class Observation(object):
    """Observation class."""

    def __init__(self, obstime, lon, lat, value, elev=np.nan, stid="NA", varname=None):
        """Construct an observation.

        Args:
            obstime (_type_): _description_
            lon (_type_): _description_
            lat (_type_): _description_
            value (_type_): _description_
            elev (_type_, optional): _description_. Defaults to np.nan.
            stid (str, optional): _description_. Defaults to "NA".
            varname (_type_, optional): _description_. Defaults to None.

        """
        self.obstime = obstime
        self.lon = float(lon)
        self.lat = float(lat)
        self.stid = stid
        self.elev = float(elev)
        self.value = float(value)
        self.varname = varname

    def print_obs(self):
        """Print observation."""
        print(
            "observation: ",
            self.obstime,
            self.lon,
            self.lat,
            self.stid,
            self.value,
            self.elev,
        )

    @staticmethod
    def vectors2obs(obstime, lon, lat, stid, elev, value, varname):
        """Convert vectors to observations.

        Args:
            obstime (_type_): _description_
            lon (_type_): _description_
            lat (_type_): _description_
            stid (_type_): _description_
            elev (_type_): _description_
            value (_type_): _description_
            varname (_type_): _description_

        Returns:
            Observation: Observation object.
        """
        return Observation(
            obstime, lon, lat, value, elev=elev, varname=varname, stid=stid
        )

    @staticmethod
    def obs2vectors(my_obs):
        """Convert observations to vectors.

        Args:
            my_obs (_type_): _description_

        Returns:
            _type_: _description_
        """
        return (
            my_obs.obstime,
            my_obs.lon,
            my_obs.lat,
            my_obs.stid,
            my_obs.elev,
            my_obs.value,
            my_obs.varname,
        )

    @staticmethod
    def format_lon(lon):
        """Format longitude."""
        lon = f"{float(lon):10.5f}"
        return lon

    @staticmethod
    def format_lat(lat):
        """Format latitude."""
        lat = f"{float(lat):10.5f}"
        return lat

    @staticmethod
    def get_pos_from_stid(filename, stids):
        """Get pos from station ID.

        Args:
            filename (_type_): _description_
            stids (_type_): _description_

        Raises:
            Exception: _description_

        Returns:
            _type_: _description_

        """
        lons = []
        lats = []
        with open(filename, mode="r", encoding="utf-8") as file_handler:
            ids_from_file = json.load(file_handler)
        for stid in stids:
            found = False
            for stid1 in ids_from_file:
                if stid == stid1:
                    found = True
                    lon = float(ids_from_file[stid1]["lon"])
                    lat = float(ids_from_file[stid1]["lat"])
                    lons.append(lon)
                    lats.append(lat)
            if not found:
                raise Exception(
                    "Could not find station id " + stid + " in file " + filename
                )
        return lons, lats

    @staticmethod
    def get_stid_from_stationlist(stationlist, lons, lats):
        """Get station ID from station list.

        Args:
            stationlist (str): Filename of station list
            lons (list): Longitudes
            lats (list): Latitudes

        Returns:
            list: Station IDs

        """
        index_pos = {}
        with open(stationlist, mode="r", encoding="utf-8") as file_handler:
            ids_from_file = json.load(file_handler)
        for stid in ids_from_file:
            lon = ids_from_file[stid]["lon"]
            lat = ids_from_file[stid]["lat"]
            pos = Observation.format_lon(lon) + ":" + Observation.format_lat(lat)
            index_pos.update({pos: stid})

        stids = []
        for i, lon in enumerate(lons):
            lat = lats[i]
            pos = Observation.format_lon(lon) + ":" + Observation.format_lat(lat)
            if pos in index_pos:
                stids.append(index_pos[pos])
            else:
                stids.append("NA")
        return stids



"""Converter and read data."""
import copy
import logging
from abc import ABCMeta, abstractmethod

import numpy as np

from .interpolation import fill_field
from .util import deep_update
from .variable import Variable


class ReadData(object):
    """Read data class."""

    __metaclass__ = ABCMeta

    def __init__(self, geo, var_name):
        """Construct readData object."""
        self.geo = geo
        self.var_name = var_name

    @abstractmethod
    def read_time_step(self, validtime, cache):
        """To be implemented."""
        raise NotImplementedError(
            "users must define read_time_step to use this base class"
        )

    @abstractmethod
    def print_info(self):
        """To be implemented."""
        raise NotImplementedError(
            "users must define read_time_step to use this base class"
        )


# Direct data can be ead with this class with converter = None
class ConvertedInput(ReadData):
    """Converted input.

    Args:
        geo (surfex.Geo): Surfex geometry
        var_name (str): Variable name
        converter (surfex.Converter): Converter
    """

    def __init__(self, geo, var_name, converter):
        """Construct converted input.

        Args:
            geo (_type_): _description_
            var_name (_type_): _description_
            converter (_type_): _description_

        """
        ReadData.__init__(self, geo, var_name)
        self.converter = converter

    def read_time_step(self, validtime, cache):
        """Read time  step.

        Args:
            validtime (_type_): _description_
            cache (_type_): _description_

        Returns:
            _type_: _description_

        """
        field = self.converter.read_time_step(self.geo, validtime, cache)
        # Preserve positive values for precipitation
        # TODO
        """
        if self.var_name == "RAIN" or self.var_name == "SNOW":
            field[field < 0.] = 0.
        """
        return field

    def print_info(self):
        """Print info."""
        return self.converter.print_info()


class ConstantValue(ReadData):
    """Constant value converter."""

    def __init__(self, geo, var_name, var_dict):
        """Set a field to a constant value.

        Args:
            geo (surfex.Geo): Geometry object
            var_name (str): Variable name
            var_dict (dict): Variable definition

        Raises:
            RuntimeError: Constant value must have a value!

        """
        ReadData.__init__(self, geo, var_name)
        self.var_dict = var_dict
        if "value" in self.var_dict:
            self.value = self.var_dict["value"]
        else:
            raise RuntimeError("Constant value must have a value!")

    def read_time_step(self, validtime, cache):
        """Read time step.

        Args:
            validtime (_type_): _description_
            cache (_type_): _description_

        Returns:
            _type_: _description_

        """
        field = np.array([float(i) for i in range(0, self.geo.npoints)])
        field.fill(self.value)
        return field

    def print_info(self):
        """Print info."""
        return self.var_name


#######################################################
#######################################################


class Converter(object):
    """Converter.

    Main interface to read a field is done through a converter.
    The converter name is default "None" to read a plain field without any conversion.

    """

    def __init__(self, name, initial_time, defs, conf, fileformat):
        """Initialize the converter.

        Args:
            name (str): Converter name
            initial_time (as_datetime): The valid time you want to read
            defs (dict): A dictionary defining the variables
            conf (dict): A dictionary defining the converter
            fileformat (str): File format

        Raises:
            KeyError: Missing definitions
            NotImplementedError: Converter not implemented

        """
        self.name = name
        self.initial_time = initial_time

        logging.debug("Converter name: %s", self.name)
        logging.debug("Converter config: %s", conf)
        if self.name not in conf:
            logging.debug("conf: %s", conf)
            raise KeyError(self.name + " is missing in converter definition")

        if self.name == "none" or self.name == "analysis":
            self.var = self.create_variable(fileformat, defs, conf[self.name])
        elif name == "rh2q":
            self.r_h = self.create_variable(fileformat, defs, conf[self.name]["rh"])
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
            self.pres = self.create_variable(fileformat, defs, conf[self.name]["p"])
        elif name == "mslp2ps":
            self.altitude = self.create_variable(
                fileformat, defs, conf[self.name]["altitude"]
            )
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
            self.pres = self.create_variable(fileformat, defs, conf[self.name]["mslp"])
        elif name == "rh2q_mslp":
            self.r_h = self.create_variable(fileformat, defs, conf[self.name]["rh"])
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
            self.altitude = self.create_variable(
                fileformat, defs, conf[self.name]["altitude"]
            )
            self.pres = self.create_variable(fileformat, defs, conf[self.name]["mslp"])
        elif name == "windspeed" or name == "winddir":
            self.x_wind = self.create_variable(fileformat, defs, conf[self.name]["x"])
            self.y_wind = self.create_variable(fileformat, defs, conf[self.name]["y"])
        elif name == "totalprec":
            self.totalprec = self.create_variable(
                fileformat, defs, conf[self.name]["totalprec"]
            )
            self.snow = self.create_variable(fileformat, defs, conf[self.name]["snow"])
        elif name == "calcsnow":
            self.totalprec = self.create_variable(
                fileformat, defs, conf[self.name]["totalprec"]
            )
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
        elif name == "calcrain":
            self.totalprec = self.create_variable(
                fileformat, defs, conf[self.name]["totalprec"]
            )
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
        elif name == "snowplusgraupel":
            self.snow = self.create_variable(fileformat, defs, conf[self.name]["snow"])
            self.graupel = self.create_variable(
                fileformat, defs, conf[self.name]["graupel"]
            )
        elif name == "phi2m":
            self.phi = self.create_variable(fileformat, defs, conf[self.name]["phi"])
        elif self.name == "swe2sd":
            self.swe = self.create_variable(fileformat, defs, conf[self.name]["swe"])
        elif self.name == "sweclim":
            self.swe = self.create_variable(fileformat, defs, conf[self.name]["swe"])
        elif self.name == "sea2land":
            self.sea = self.create_variable(fileformat, defs, conf[self.name]["sea"])
        elif self.name == "tap":
            self.tap1 = self.create_variable(fileformat, defs, conf[self.name]["tap1"])
            self.tap2 = self.create_variable(fileformat, defs, conf[self.name]["tap2"])
        elif self.name == "rhp":
            self.rhp1 = self.create_variable(fileformat, defs, conf[self.name]["rhp1"])
            self.rhp2 = self.create_variable(fileformat, defs, conf[self.name]["rhp2"])
        elif self.name == "sdp":
            self.sdp1 = self.create_variable(fileformat, defs, conf[self.name]["sdp1"])
            self.sdp2 = self.create_variable(fileformat, defs, conf[self.name]["sdp2"])
        elif self.name == "smp":
            self.smp1 = self.create_variable(fileformat, defs, conf[self.name]["smp1"])
            self.smp2 = self.create_variable(fileformat, defs, conf[self.name]["smp2"])
        elif self.name == "nature_town":
            self.nature_fraction = self.create_variable(
                fileformat, defs, conf[self.name]["nature_fraction"]
            )
            self.town_fraction = self.create_variable(
                fileformat, defs, conf[self.name]["town_fraction"]
            )
        elif self.name == "cloud_base":
            self.cloud_base = self.create_variable(
                fileformat, defs, conf[self.name]["cloud_base"]
            )
        else:
            raise NotImplementedError("Converter " + self.name + " not implemented")

        logging.debug("Constructed the converter %s", self.name)

    def print_info(self):
        """Print info."""
        return f"name: {self.name}"

    def create_variable(self, fileformat, defs, var_dict):
        """Create a variable.

        Args:
            fileformat (str): Fileformat
            defs (dict): defs
            var_dict (dict): Variable dictionary

        Raises:
            RuntimeError: Variable is not set

        Returns:
            field: The read field

        """
        # Finally we can merge the variable with the default settings
        # Create deep copies not to inherit between variables
        defs = copy.deepcopy(defs)
        var_dict = copy.deepcopy(var_dict)
        if var_dict is None:
            raise RuntimeError("Variable is not set")
        merged_dict = deep_update(defs, var_dict)

        var = Variable(fileformat, merged_dict, self.initial_time)

        logging.debug(var.print_variable_info())
        return var

    @staticmethod
    def mslp2ps(mslp, altitude, temp):
        """Calcaulate ps from mslp.

        Args:
            mslp (np.ndarray): Mean sea level pressure
            altitude (np.ndarray): Altitude
            temp (np.ndarray): Temperature

        Returns:
            np.ndarray: Surface pressure

        """
        gravity = 9.81
        dry_air = 287.0

        pres = np.multiply(
            mslp,
            np.exp(
                np.divide(np.multiply(-altitude, gravity), np.multiply(dry_air, temp))
            ),
        )

        return pres

    def read_time_step(self, geo, validtime, cache):
        """Read time step.

        Args:
            geo (Geo): Geometry_
            validtime (as_datetime): Validtime
            cache (Cache): Cache

        Raises:
            KeyError:Could not found climatological mean for month
            NotImplementedError: _description_

        Returns:
            field (np.ndarray): Read and converted field

        """
        gravity = 9.81
        field = None
        # Specific reading for each converter
        if self.name == "none" or self.name == "analysis":
            field = self.var.read_variable(geo, validtime, cache)
        elif self.name == "windspeed" or self.name == "winddir":
            field_x = self.x_wind.read_variable(geo, validtime, cache)
            field_y = self.y_wind.read_variable(geo, validtime, cache)
            if self.name == "windspeed":
                field = np.sqrt(np.square(field_x) + np.square(field_y))
                np.where(field < 0.005, field, 0)
            elif self.name == "winddir":
                field = np.mod(90 - np.rad2deg(np.arctan2(field_y, field_x)), 360)

        elif self.name == "rh2q" or self.name == "rh2q_mslp":
            """
            ZES = 6.112 * exp((17.67 * (ZT - 273.15)) / ((ZT - 273.15) + 243.5))
            ZE = ZRH * ZES
            ZRATIO = 0.622 * ZE / (ZPRES / 100.)
            RH2Q = 1. / (1. / ZRATIO + 1.)
            """
            field_r_h = self.r_h.read_variable(geo, validtime, cache)  # %
            field_temp = self.temp.read_variable(geo, validtime, cache)  # In K
            field_pres = self.pres.read_variable(geo, validtime, cache)  # In Pa
            if self.name == "rh2q_mslp":
                field_altitude = self.altitude.read_variable(
                    geo, validtime, cache
                )  # In m
                field_pres = self.mslp2ps(field_pres, field_altitude, field_temp)

            field_p_mb = np.divide(field_pres, 100.0)
            field_t_c = np.subtract(field_temp, 273.15)

            exp = np.divide(np.multiply(17.67, field_t_c), np.add(field_t_c, 243.5))
            esat = np.multiply(6.112, np.exp(exp))
            field = np.divide(np.multiply(0.622, field_r_h / 100.0) * esat, field_p_mb)

        elif self.name == "mslp2ps":
            field_pres = self.pres.read_variable(geo, validtime, cache)  # In Pa
            field_temp = self.temp.read_variable(geo, validtime, cache)  # In K
            field_altitude = self.altitude.read_variable(geo, validtime, cache)  # In m
            field = self.mslp2ps(field_pres, field_altitude, field_temp)
        elif self.name == "totalprec":
            field_totalprec = self.totalprec.read_variable(geo, validtime, cache)
            field_snow = self.snow.read_variable(geo, validtime, cache)
            field = np.subtract(field_totalprec, field_snow)
            if any(field[field < 0.0]):
                logging.info("Set negative rain values to zero")
                field[field < 0.0] = 0

        elif self.name == "calcrain":
            field_totalprec = self.totalprec.read_variable(geo, validtime, cache)
            field_t = self.temp.read_variable(geo, validtime, cache)
            field = field_totalprec
            field[field_t <= 274.16] = 0
        elif self.name == "calcsnow":
            field_totalprec = self.totalprec.read_variable(geo, validtime, cache)
            field_t = self.temp.read_variable(geo, validtime, cache)  # In K
            field = field_totalprec
            field[field_t > 274.16] = 0
        elif self.name == "snowplusgraupel":
            field = self.snow.read_variable(geo, validtime, cache)
            field += self.graupel.read_variable(geo, validtime, cache)
        elif self.name == "phi2m":
            field = self.phi.read_variable(geo, validtime, cache)
            field = np.divide(field, gravity)
            field[(field < 0)] = 0.0
        elif self.name == "swe2sd":
            field = self.swe.read_variable(geo, validtime, cache)
            rho = self.swe.read_variable(geo, validtime, cache)
            field = np.divide(field, rho)
        elif self.name == "sweclim":
            field = self.swe.read_variable(geo, validtime, cache)
            rhoclim = {
                "01": 222.0,
                "02": 233.0,
                "03": 240.0,
                "04": 278.0,
                "05": 212.0,
                "06": 312.0,
                "07": 312.0,
                "08": 143.0,
                "09": 143.0,
                "10": 161.0,
                "11": 182.0,
                "12": 213.0,
            }
            month = validtime.strftime("%m")
            if month in rhoclim:
                field = np.divide(field, rhoclim[month])
            else:
                raise KeyError(
                    "Could not found climatological mean for month " + str(month)
                )
        elif self.name == "sea2land":
            field = self.sea.read_variable(geo, validtime, cache)
            field = np.subtract(1, field)
        elif self.name == "tap":
            tap1 = self.tap1.read_variable(geo, validtime, cache)
            tap2 = self.tap2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(tap1), tap2, tap1)
        elif self.name == "rhp":
            rhp1 = self.rhp1.read_variable(geo, validtime, cache)
            rhp2 = self.rhp2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(rhp1), rhp2, rhp1)
        elif self.name == "sdp":
            sdp1 = self.sdp1.read_variable(geo, validtime, cache)
            sdp2 = self.sdp2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(sdp1), sdp2, sdp1)
        elif self.name == "smp":
            smp1 = self.smp1.read_variable(geo, validtime, cache)
            smp2 = self.smp2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(smp1), smp2, smp1)
        elif self.name == "nature_town":
            nature = self.nature_fraction.read_variable(geo, validtime, cache)
            town = self.town_fraction.read_variable(geo, validtime, cache)
            field = np.add(nature, town)
            field[field > 1] = 1.0
        elif self.name == "cloud_base":
            logging.info("Converter cloud_base")

            field = self.cloud_base.read_variable(geo, validtime, cache)
            field_2d = field.reshape(geo.nlons, geo.nlats)

            iteration = 0
            while np.any(np.isnan(field_2d)):
                logging.debug("Filling cloud base")
                field_2d, nans = fill_field(field_2d, geo, radius=3)
                iteration = iteration + 1
                logging.debug("Iteration %s NaNs: %s", iteration, nans)

            # Reshape back to 1D
            field = field_2d.reshape(geo.nlons * geo.nlats)

        else:
            raise NotImplementedError("Converter " + self.name + " not implemented")
        return field



"""Grib treatment."""
import logging

import numpy as np
import pyproj

try:
    import eccodes
    import gribapi
except ImportError:
    eccodes = None
    gribapi = None
except RuntimeError:
    eccodes = None
    gribapi = None
# Needed in Python 3.5
except:  # noqa
    eccodes = None
    gribapi = None


from .geo import ConfProj, Geo, LonLatReg
from .interpolation import Interpolation


class Grib(object):
    """Grib class."""

    def __init__(self, fname):
        """Construct grib object.

        Args:
            fname (str): File name.

        """
        self.fname = fname
        self.projection = None
        self.lons = None
        self.lats = None
        self.nearest = None
        self.linear = None
        logging.debug("Grib constructor")

    def field(self, gribvar, time):
        """Read field in grib file.

        Args:
            gribvar (GribVariable1/2): Grib variable to read.
            time (datetime.datetime): Valid time to read.

        Returns:
            np.ndarray: Field

        Raises:
            NotImplementedError: NotImplementedError
            RuntimeError: If eccodes not available

        """
        if eccodes is None:
            raise RuntimeError("eccodes not found. Needed for reading grib files")

        logging.debug("Look for %s", gribvar.generate_grib_id())

        field = None
        geo_out = None
        with open(self.fname, mode="rb") as file_handler:
            while 1:
                gid = eccodes.codes_grib_new_from_file(file_handler)

                if gid is None:
                    logging.warning("Could not find key")
                    gribvar.print_keys()
                    file_handler.close()
                    return field, geo_out
                else:
                    if gribvar.matches(gid):
                        values = self.read_field_in_message(gid, time)
                        logging.debug("read values = %s", values)

                        grid_type = str(eccodes.codes_get(gid, "gridType"))
                        logging.debug("grid_type=%s", grid_type)
                        if grid_type.lower() == "rotated_ll":
                            geo_keys = [
                                "Ni",
                                "Nj",
                                "latitudeOfFirstGridPointInDegrees",
                                "longitudeOfFirstGridPointInDegrees",
                                "latitudeOfLastGridPointInDegrees",
                                "longitudeOfLastGridPointInDegrees",
                                "iDirectionIncrementInDegrees",
                                "jDirectionIncrementInDegrees",
                                "latitudeOfSouthernPoleInDegrees",
                                "longitudeOfSouthernPoleInDegrees",
                                "iScansNegatively",
                                "jScansPositively",
                            ]
                            geo_info = self.read_geo_info(gid, geo_keys)

                            n_x = geo_info["Ni"]
                            n_y = geo_info["Nj"]

                            ll_lon = geo_info["longitudeOfFirstGridPointInDegrees"]
                            ll_lat = geo_info["latitudeOfFirstGridPointInDegrees"]
                            dlon = geo_info["iDirectionIncrementInDegrees"]
                            dlat = geo_info["jDirectionIncrementInDegrees"]
                            sp_lon = geo_info["longitudeOfSouthernPoleInDegrees"]
                            iscan = geo_info["iScansNegatively"]
                            jscan = geo_info["jScansPositively"]
                            if sp_lon < -180.0:
                                sp_lon = sp_lon + 360.0
                            elif sp_lon > 180.0:
                                sp_lon = sp_lon - 360.0
                            sp_lat = -1 * geo_info["latitudeOfSouthernPoleInDegrees"]
                            earth = 6.371229e6

                            proj_string = (
                                f"+proj=ob_tran +o_proj=longlat +o_lat_p={sp_lat}"
                                f" +R={str(earth)} +no_defs"
                            )
                            logging.info(proj_string)
                            logging.info("ll_lon=%s ll_lat=%s", ll_lon, ll_lat)
                            logging.info("polon=%s polat=%s", sp_lon, sp_lat)
                            logging.info("dlon=%s dlat=%s", dlon, dlat)
                            logging.info("iscan=%s jscan=%s", iscan, jscan)
                            proj = pyproj.CRS.from_string(proj_string)
                            wgs84 = pyproj.CRS.from_string("EPSG:4326")

                            lons = []
                            for i in range(0, n_x):
                                if int(iscan) == 1:
                                    lon = ll_lon - (float(i) * dlon)
                                else:
                                    lon = ll_lon + (float(i) * dlon)
                                if lon < -180.0:
                                    lon = lon + 360.0
                                elif lon > 180.0:
                                    lon = lon - 360.0
                                lons.append(lon)
                            lats = []
                            for j in range(0, n_y):
                                if int(jscan) == 1:
                                    lat = ll_lat + (float(j) * dlat)
                                else:
                                    lat = ll_lat - (float(j) * dlat)
                                if lat > 90.0:
                                    lat = lat - 90.0
                                elif lat < -90.0:
                                    lat = lat + 90.0
                                lats.append(lat)

                            lons = np.array(lons)
                            lats = np.array(lats)
                            longitudes, latitudes = np.meshgrid(lons, lats, indexing="ij")
                            lons, lats = pyproj.Transformer.from_crs(
                                proj, wgs84, always_xy=True
                            ).transform(longitudes, latitudes)
                            lons = lons + sp_lon

                            field = np.reshape(values, [n_x, n_y], order="F")
                            if geo_out is None:
                                geo_out = Geo(lons, lats)

                        elif grid_type.lower() == "regular_ll":
                            geo_keys = [
                                "Ni",
                                "Nj",
                                "latitudeOfFirstGridPointInDegrees",
                                "longitudeOfFirstGridPointInDegrees",
                                "latitudeOfLastGridPointInDegrees",
                                "longitudeOfLastGridPointInDegrees",
                                "iDirectionIncrementInDegrees",
                                "jDirectionIncrementInDegrees",
                            ]
                            geo_info = self.read_geo_info(gid, geo_keys)
                            n_x = geo_info["Ni"]
                            n_y = geo_info["Nj"]
                            lon0 = geo_info["longitudeOfFirstGridPointInDegrees"]
                            lat0 = geo_info["latitudeOfFirstGridPointInDegrees"]
                            d_x = geo_info["iDirectionIncrementInDegrees"]
                            d_y = geo_info["jDirectionIncrementInDegrees"]
                            lons = []
                            lats = []
                            for i in range(0, n_x):
                                lons.append(lon0 + (float(i) * d_x))
                            for j in range(0, n_y):
                                lats.append(lat0 - (float(j) * d_y))
                            lon1 = lons[-1]
                            lat1 = lats[-1]
                            lons = np.array(lons)
                            lats = np.array(lats)
                            lons, lats = np.meshgrid(lons, lats)
                            field = np.reshape(values, [n_x, n_y], order="F")

                            if geo_out is None:
                                domain = {
                                    "nam_lonlat_reg": {
                                        "xlonmin": lon0,
                                        "xlonmax": lon1,
                                        "xlatmin": lat0,
                                        "xlatmax": lat1,
                                        "nlon": n_x,
                                        "nlat": n_y,
                                    }
                                }
                                geo_out = LonLatReg(domain)

                        elif grid_type.lower() == "lambert":
                            geo_keys = [
                                "Nx",
                                "Ny",
                                "latitudeOfFirstGridPointInDegrees",
                                "longitudeOfFirstGridPointInDegrees",
                                "LoVInDegrees",
                                "DxInMetres",
                                "DyInMetres",
                                "iScansNegatively",
                                "jScansPositively",
                                "jPointsAreConsecutive",
                                "Latin1InDegrees",
                                "LaDInDegrees",
                                "Latin2InDegrees",
                                "latitudeOfSouthernPoleInDegrees",
                                "longitudeOfSouthernPoleInDegrees",
                            ]
                            geo_info = self.read_geo_info(gid, geo_keys)

                            n_x = geo_info["Nx"]
                            n_y = geo_info["Ny"]

                            lon0 = geo_info["LoVInDegrees"]
                            lat0 = geo_info["LaDInDegrees"]
                            ll_lon = geo_info["longitudeOfFirstGridPointInDegrees"]
                            ll_lat = geo_info["latitudeOfFirstGridPointInDegrees"]
                            d_x = geo_info["DxInMetres"]
                            d_y = geo_info["DyInMetres"]

                            earth = 6.37122e6
                            proj_string = (
                                f"+proj=lcc +lat_0={str(lat0)} +lon_0={str(lon0)} "
                                f"+lat_1={str(lat0)} +lat_2={str(lat0)} "
                                f"+units=m +no_defs +R={str(earth)}"
                            )

                            proj = pyproj.CRS.from_string(proj_string)
                            wgs84 = pyproj.CRS.from_string("EPSG:4326")
                            x_0, y_0 = pyproj.Transformer.from_crs(
                                wgs84, proj, always_xy=True
                            ).transform(ll_lon, ll_lat)
                            x_c = x_0 + 0.5 * (n_x - 1) * d_x
                            y_c = y_0 + 0.5 * (n_y - 1) * d_y
                            lonc, latc = pyproj.Transformer.from_crs(
                                proj, wgs84, always_xy=True
                            ).transform(x_c, y_c)

                            # TODO we should investigate scan angle and if done correctly
                            # order should probaly be "C" and not "F"
                            field = np.reshape(values, [n_x, n_y], order="F")

                            if geo_out is None:
                                domain = {
                                    "nam_conf_proj": {"xlon0": lon0, "xlat0": lat0},
                                    "nam_conf_proj_grid": {
                                        "xloncen": lonc,
                                        "xlatcen": latc,
                                        "nimax": n_x,
                                        "njmax": n_y,
                                        "xdx": d_x,
                                        "xdy": d_y,
                                        "ilone": 0,
                                        "ilate": 0,
                                    },
                                }
                                geo_out = ConfProj(domain)
                        else:
                            raise NotImplementedError(
                                str(grid_type) + " not implemented yet!"
                            )

                        eccodes.codes_release(gid)
                        if geo_out is None:
                            raise RuntimeError("No geometry is found in file")

                        return field, geo_out
                    eccodes.codes_release(gid)

    @staticmethod
    def read_geo_info(gid, keys):
        """Set geo keys in dict.

        Args:
            gid (int): grib id
            keys (dict): Geometry keys

        Returns:
            dict: Geometry

        """
        geo_dict = {}
        for key in keys:
            try:
                logging.debug("  %s: %s", key, eccodes.codes_get(gid, key))
                geo_dict.update({key: eccodes.codes_get(gid, key)})
            except eccodes.KeyValueNotFoundError as err:
                logging.debug('  Key="%s" was not found: %s', key, err.msg)
            except eccodes.CodesInternalError as err:
                logging.error('Error with key="%s" : %s', key, err.msg)
        return geo_dict

    @staticmethod
    def read_field_in_message(gid, time):
        """Get field.

        Args:
            gid (int): grib id
            time(str): time

        Returns:
            Field: Field object

        """
        try:
            logging.debug(
                "There are %d values, average is %f, min is %f, max is %f",
                eccodes.codes_get_size(gid, "values"),
                eccodes.codes_get(gid, "average"),
                eccodes.codes_get(gid, "min"),
                eccodes.codes_get(gid, "max"),
            )
            field = eccodes.codes_get_values(gid)

            # TODO Check time consistency
            logging.info("Grib record found is hopefullly valid for time %s", str(time))

            has_bitmap = 0
            try:
                has_bitmap = int(eccodes.codes_get(gid, "bitmapPresent"))
            except eccodes.KeyValueNotFoundError as err:
                logging.debug('  Key="bitmapPresent" was not found: %s', err.msg)
            except eccodes.CodesInternalError as err:
                logging.error('Error with key="bitmapPresent" : %s', err.msg)
            if has_bitmap == 1:
                missing_value = eccodes.codes_get(gid, "missingValue")
                field[field == missing_value] = np.nan
            return field
        except eccodes.KeyValueNotFoundError as err:
            logging.debug('  Key="missingValue" was not found: %s', err.msg)
        except eccodes.CodesInternalError as err:
            logging.error('Error with key="missingValue" : %s', err.msg)
        return None

    def points(self, gribvar, geo, validtime=None, interpolation="bilinear"):
        """Read a 2-D field and interpolates it to requested positions.

        Args:
            gribvar (GribVariable1/2): Grib variable
            geo (surfex.Geo): Surfex geometry
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            interpolation (str, optional): Interpolation method. Defaults to "bilinear".

        Returns:
             np.array: vector with interpolated values

        """
        field, geo_in = self.field(gribvar, validtime)
        interpolator = Interpolation(interpolation, geo_in, geo)
        field = interpolator.interpolate(field)
        return field, interpolator


class Grib1Variable:
    """Grib1 variable."""

    def __init__(self, par, typ, level, tri):
        """Construct grib1 variable.

        Args:
            par (int): IndicatorOfParameter
            typ (int): TypeOfLevel
            level (int): NumberOFLevel
            tri (int): TimeRangeIndicator

        """
        self.version = 1
        self.par = int(par)
        self.typ = int(typ)
        self.level = int(level)
        self.tri = int(tri)

    def is_accumulated(self):
        """Check if accumulated."""
        if self.tri == 4:
            logging.debug("Value is accumulated")
            return True
        else:
            return False

    def matches(self, gid):
        """Check if matchees.

        Args:
            gid (_type_): _description_

        Returns:
            bool: True if found

        Raises:
            RuntimeError: If eccodes not found

        """
        if eccodes is None:
            raise RuntimeError("eccodes not found. Needed for reading grib files")

        version = int(eccodes.codes_get(gid, "editionNumber"))
        if version == 1:
            par = int(eccodes.codes_get_long(gid, "indicatorOfParameter"))
            lev = int(eccodes.codes_get_long(gid, "level"))
            typ = int(eccodes.codes_get_long(gid, "levelType"))
            tri = int(eccodes.codes_get_long(gid, "timeRangeIndicator"))

            logging.debug("Checking grib1 record: %s %s %s %s", par, typ, lev, tri)
            logging.debug(self.generate_grib_id())
            if (
                self.par == par
                and self.level == lev
                and self.typ == typ
                and self.tri == tri
            ):
                logging.debug(
                    "Found matching grib1 record: %s %s %s %s", par, lev, typ, tri
                )
                return True
            else:
                return False
        else:
            logging.warning("Record is not grib1")
            return False

    def print_keys(self):
        """Print keys."""
        print("\n")
        print("Version:", self.version)
        print("indicatorOfParameter:", self.par)
        print("levelType", self.typ)
        print("level:", self.level)
        print("timeRangeIndicator:", self.tri)

    def generate_grib_id(self):
        """Generate grib1 ID."""
        par = self.par
        typ = self.typ
        level = self.level
        tri = self.tri
        return f":grib1:{str(par)}:{str(typ)}:{str(level)}:{str(tri)}"


class Grib2Variable(object):
    """Grib2 variable."""

    def __init__(self, discipline, pca, pnr, typ, lev, tsp=-1):
        """Construct a grib2 variable.

        Args:
            discipline (int): discipline
            pca (int): parameterCatergory
            pnr (int): parameterNumber
            typ (int): levelType
            lev (int): level
            tsp (int, optional): typeOfStatisticalProcessing. Defaults to -1.
        """
        self.version = 2
        self.discipline = int(discipline)
        self.parameter_category = int(pca)
        self.parameter_number = int(pnr)
        self.level_type = int(typ)
        self.level = int(lev)
        self.type_of_statistical_processing = int(tsp)

    def matches(self, gid):
        """Check if matches.

        Args:
            gid (int): Grib ID

        Raises:
            ModuleNotFoundError: If not eccodes found

        Returns:
            bool: True if matches
        """
        if eccodes is None:
            raise ModuleNotFoundError("eccodes not found. Needed for reading grib files")

        version = int(eccodes.codes_get(gid, "editionNumber"))
        if version == 2:
            discipline = int(eccodes.codes_get(gid, "discipline"))
            parameter_category = int(eccodes.codes_get(gid, "parameterCategory"))
            parameter_number = int(eccodes.codes_get(gid, "parameterNumber"))
            level_type = int(eccodes.codes_get_long(gid, "levelType"))
            level = int(eccodes.codes_get(gid, "level"))
            try:
                type_of_statistical_processing = eccodes.codes_get(
                    gid, "typeOfStatisticalProcessing"
                )
                type_of_statistical_processing = int(type_of_statistical_processing)
            except gribapi.errors.KeyValueNotFoundError:
                type_of_statistical_processing = -1

            logging.debug(
                "Checking grib2 record: %s %s %s %s %s %s",
                discipline,
                parameter_category,
                parameter_number,
                level_type,
                level,
                type_of_statistical_processing,
            )
            logging.debug("grib2 ID: %s", self.generate_grib_id())
            if (
                self.discipline == discipline
                and self.parameter_category == parameter_category
                and self.parameter_number == parameter_number
                and self.level_type == level_type
                and self.level == level
                and self.type_of_statistical_processing == type_of_statistical_processing
            ):
                logging.debug(
                    "Found matching grib2 record: %s %s %s %s %s",
                    discipline,
                    parameter_category,
                    parameter_number,
                    level_type,
                    level,
                )
                return True
            else:
                return False
        else:
            logging.warning("Record is not grib2")
            return False

    def is_accumulated(self):
        """Check if accumulated."""
        if self.type_of_statistical_processing == 1:
            logging.debug("Parameter is accumulated")
            return True
        else:
            return False

    def print_keys(self):
        """Print keys."""
        print("\n")
        print("Version:", self.version)
        print("discipline:", self.discipline)
        print("parameterCategory:", self.parameter_category)
        print("parameterNumber:", self.parameter_number)
        print("levelType:", self.level_type)
        print("level:", self.level)
        print("typeOfStatisticalProcessing:", self.type_of_statistical_processing)

    def generate_grib_id(self):
        """Generate grib2 ID."""
        dis = str(self.discipline)
        pca = str(self.parameter_category)
        pnr = str(self.parameter_number)
        typ = str(self.level_type)
        lev = str(self.level)
        tsp = str(self.type_of_statistical_processing)
        return f":grib2:{dis}:{pca}:{pnr}:{typ}:{lev}:{tsp}"



"""Titan."""
import abc
import json
import logging
import os

import numpy as np

try:
    import titanlib as tit
except ImportError:
    tit = None

from .datetime_utils import as_datetime, as_datetime_string
from .interpolation import ObsOperator, inside_grid
from .netcdf import read_first_guess_netcdf_file
from .observation import Observation


class QualityControl(object):
    """Quality control abstract class."""

    def __init__(self, name):
        """Construct base class."""
        self.name = name

    @abc.abstractmethod
    def set_input(self, size):
        """Set input to test.

        Args:
            size (int): Number of observations.

        Raises:
            NotImplementedError: Must be implemented by child class.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def test(self, dataset, mask, code=1):
        """The quality control test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 1.

        Raises:
            NotImplementedError: Must be implemented by child class.
        """
        raise NotImplementedError

    @staticmethod
    def set_flags(global_flags, flags, mask, code):
        """Set flags.

        Args:
            global_flags (list): Global flags
            flags (list): Test flags
            mask (list): Active data
            code (int): Code to use for flagging.

        Returns:
            list: Updated global flags.
        """
        imask = np.where((np.array(global_flags) == 0) & (np.array(list(flags)) == 1))[0]
        imask = np.intersect1d(imask, np.array(mask))
        if len(imask) > 0:
            global_flags[imask] = code

        return global_flags


class Plausibility(QualityControl):
    """Plausibilty."""

    def __init__(self, minval=None, maxval=None):
        """Construct plausibility test.

        Args:
            minval (float, optional): Default minimum value. Defaults to None.
            maxval (float, optional): Default maximum value. Defaults to None.

        """
        self.minvals = []
        self.maxvals = []
        self.def_min = minval
        self.def_max = maxval
        QualityControl.__init__(self, "plausibility")

    def set_input(self, size, minval=None, maxval=None):
        """Set input.

        Args:
            size (size): Observation data set size
            minval (float, optional): Minimum value. Defaults to None.
            maxval (float, optional): Maximum value. Defaults to None.

        Raises:
            RuntimeError: You must set minval and maxval

        """
        used_min = self.def_min
        used_max = self.def_max
        if minval is not None:
            used_min = minval
        if maxval is not None:
            used_max = maxval

        if used_min is None or used_max is None:
            raise RuntimeError("You must set minval and maxval")

        logging.debug("minval: %s", used_min)
        logging.debug("maxval: %s", used_max)

        minvals = []
        maxvals = []
        for __ in range(0, size):
            minvals.append(used_min)
            maxvals.append(used_max)

        self.minvals = self.minvals + minvals
        self.maxvals = self.maxvals + maxvals

    def test(self, dataset, mask, code=102):
        """Do the test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 102.

        Raises:
            ModuleNotFoundError: titanlib was not loaded properly

        Returns:
            global_flags(list): Global flags.

        """
        if tit is None:
            raise ModuleNotFoundError("titanlib was not loaded properly")

        minvals = []
        maxvals = []
        values = []
        for i, mask_ind in enumerate(mask):
            minvals.append(self.minvals[i])
            maxvals.append(self.maxvals[i])
            values.append(dataset.values[mask_ind])

        global_flags = dataset.flags
        flags = tit.range_check(values, minvals, maxvals)

        for i, mask_ind in enumerate(mask):
            if global_flags[mask_ind] == 0 and flags[i] == 1:
                global_flags[mask_ind] = code

        for i, mask_ind in enumerate(mask):
            logging.debug(
                "test=%s i=%s m_i=%s value(m_i)=%s flag(i)=%s global_flag(m_i)=%s",
                self.name,
                i,
                mask_ind,
                dataset.values[mask_ind],
                flags[i],
                global_flags[mask_ind],
            )

        return global_flags


class FirstGuess(QualityControl):
    """First guess check."""

    def __init__(
        self,
        geo_in,
        fg_field,
        negdiff=None,
        posdiff=None,
        max_distance=5000,
        operator="bilinear",
    ):
        """Construct first guess QC check.

        Args:
            geo_in (surfex.Geo): Surfex geometry.
            fg_field (np.ndarray): First guess field.
            negdiff (float, optional): Negative difference. Defaults to None.
            posdiff (float, optional): Positive difference. Defaults to None.
            max_distance (int, optional): Max distance from grid border. Defaults to 5000.
            operator (str, optional): Interpolation operator. Defaults to "bilinear".

        """
        self.geo_in = geo_in
        self.fg_field = fg_field
        self.def_negdiff = negdiff
        self.def_posdiff = posdiff
        self.negdiff = []
        self.posdiff = []
        self.operator = operator
        self.max_distance = max_distance
        QualityControl.__init__(self, "firstguess")

    def set_input(self, size, posdiff=None, negdiff=None):
        """Set input.

        Args:
            size (int): Observation data set size
            posdiff (float, optional): Positive diff. Defaults to None.
            negdiff (float, optional): Negative diff. Defaults to None.

        Raises:
            RuntimeError: You must set negdiff and posdiff

        """
        used_negdiff = self.def_negdiff
        used_posdiff = self.def_posdiff

        if posdiff is not None:
            used_posdiff = posdiff

        if negdiff is not None:
            used_negdiff = negdiff

        if used_negdiff is None or used_posdiff is None:
            raise RuntimeError("You must set negdiff and posdiff")

        logging.debug("posdiff: %s", used_posdiff)
        logging.debug("negdiff: %s", used_negdiff)

        minvals = []
        maxvals = []
        for __ in range(0, size):
            minvals.append(used_negdiff)
            maxvals.append(used_posdiff)

        self.negdiff = self.negdiff + minvals
        self.posdiff = self.posdiff + maxvals

    def test(self, dataset, mask, code=108):
        """Do the test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 108.

        Returns:
            global_flags(list): Global flags.

        Raises:
            ModuleNotFoundError: titanlib was not loaded properly

        """
        if tit is None:
            raise ModuleNotFoundError("titanlib was not loaded properly")

        fg_operator = ObsOperator(
            self.operator,
            self.geo_in,
            dataset,
            self.fg_field,
            max_distance=self.max_distance,
        )
        fg_vals = fg_operator.get_obs_value()
        minvals = []
        maxvals = []
        values = []
        for obs, mask_ind in enumerate(mask):
            minval = fg_vals[mask_ind] - self.negdiff[obs]
            maxval = fg_vals[mask_ind] + self.posdiff[obs]
            minvals.append(minval)
            maxvals.append(maxval)
            values.append(dataset.values[mask_ind])

        flags = tit.range_check(values, minvals, maxvals)

        global_flags = dataset.flags
        for i, mask_ind in enumerate(mask):
            if fg_operator.is_in_grid(mask[i]):
                if int(global_flags[mask_ind]) == 0 and int(flags[i]) == 1:
                    global_flags[mask_ind] = code
            else:
                global_flags[mask[i]] = 199
        for i, mask_ind in enumerate(mask):
            logging.debug(
                "test=%s i=%s m_i=%s lon(m_i)=%s lat(m_i)=%s min_val(i)=%s value(i)=%s "
                "max_val(i)=%s flag(i)=%s global_flag(m_i)=%s fg_val(m_i)=%s",
                self.name,
                i,
                mask_ind,
                dataset.lons[mask_ind],
                dataset.lats[mask_ind],
                minvals[i],
                values[i],
                maxvals[i],
                flags[i],
                global_flags[mask_ind],
                fg_vals[mask_ind],
            )

        return global_flags


class Fraction(QualityControl):
    """Fraction test."""

    def __init__(
        self,
        geo_in,
        fraction_field,
        minval=None,
        maxval=None,
        max_distance=5000,
        operator="bilinear",
    ):
        """Construct fraction test.

        Args:
            geo_in (surfex.Geo): Surfex geometry
            fraction_field (np.ndarray): Field with fractions.
            minval (float, optional): Default minimum value. Defaults to None.
            maxval (float, optional): Defualt maximum value. Defaults to None.
            max_distance (int, optional): Max distance from grid border. Defaults to 5000.
            operator (str, optional): Interpolation operator. Defaults to "bilinear".

        """
        self.geo_in = geo_in
        self.fraction_field = fraction_field
        self.def_min = minval
        self.def_max = maxval
        self.min = []
        self.max = []
        self.operator = operator
        self.max_distance = max_distance
        QualityControl.__init__(self, "fraction")

    def set_input(self, size, minval=None, maxval=None):
        """Set input.

        Args:
            size (int): Observation set size.
            minval (float, optional): Minimum value. Defaults to None.
            maxval (float, optional): Maximum value. Defaults to None.

        Raises:
            RuntimeError: You must set min and max

        """
        used_min = self.def_min
        used_max = self.def_max

        if minval is not None:
            used_min = minval

        if maxval is not None:
            used_max = maxval

        if used_min is None or used_max is None:
            raise RuntimeError("You must set min and max")

        logging.debug("min: %s", used_min)
        logging.debug("max: %s", used_max)

        minvals = []
        maxvals = []
        for __ in range(0, size):
            minvals.append(used_min)
            maxvals.append(used_max)

        self.min = self.min + minvals
        self.max = self.max + maxvals

    def test(self, dataset, mask, code=151):
        """Perform test.

        Args:
            dataset (QCDataSet): QC data set.
            mask (list): Active data
            code (int, optional): Flag. Defaults to 151.

        Returns:
            global_flags(list): Global flags.

        Raises:
            ModuleNotFoundError: titanlib was not loaded properly

        """
        if tit is None:
            raise ModuleNotFoundError("titanlib was not loaded properly")

        logging.debug("Obs operator")
        fraction = ObsOperator(
            self.operator,
            self.geo_in,
            dataset,
            self.fraction_field,
            max_distance=self.max_distance,
        )

        logging.debug("get_obs_value")
        fraction_vals = fraction.get_obs_value()
        minvals = []
        maxvals = []
        values = []

        logging.debug("setup")
        for obs, mask_ind in enumerate(mask):
            minval = self.min[obs]
            maxval = self.max[obs]
            minvals.append(minval)
            maxvals.append(maxval)
            values.append(fraction_vals[mask_ind])

        minvals = np.asarray(minvals)
        maxvals = np.asarray(maxvals)
        values = np.asarray(values)
        logging.info("Do test")
        flags = tit.range_check(values, minvals, maxvals)
        logging.info("Done test")

        global_flags = dataset.flags
        for i, mask_ind in enumerate(mask):
            if fraction.is_in_grid(mask_ind):
                if int(global_flags[mask_ind]) == 0 and int(flags[i]) == 1:
                    global_flags[mask_ind] = code
            else:
                global_flags[mask[i]] = 199

        for i, mask_ind in enumerate(mask):
            logging.debug(
                "test=%s i=%s m_i=%s lon(m_i)=%s lat(m_i)=%s min_val(i)=%s value(i)=%s "
                "maxval(i)=%s flag(i)=%s global_flag(m_i)=%s fraction(m_i)=%s",
                self.name,
                i,
                mask_ind,
                dataset.lons[mask_ind],
                dataset.lats[mask_ind],
                minvals[i],
                values[i],
                maxvals[i],
                flags[i],
                global_flags[mask_ind],
                fraction_vals[mask_ind],
            )

        return global_flags


class Sct(QualityControl):
    """Spatial consistency check."""

    def __init__(
        self,
        num_min=5,
        num_max=100,
        inner_radius=50000,
        outer_radius=150000,
        num_iterations=5,
        num_min_prof=20,
        min_elev_diff=200,
        min_horizonal_scale=10000,
        vertical_scale=200,
        pos=4,
        neg=8,
        eps2=0.5,
        cmin=0.9,
        cmax=1.1,
        missing_elev_to_zero=False,
    ):
        """Construct SCT test.

        Args:
            num_min (int, optional): num_min. Defaults to 5.
            num_max (int, optional): num_max. Defaults to 100.
            inner_radius (int, optional): inner_radius. Defaults to 50000.
            outer_radius (int, optional): outer_radius. Defaults to 150000.
            num_iterations (int, optional): num_iterations. Defaults to 5.
            num_min_prof (int, optional): num_min_prof. Defaults to 20.
            min_elev_diff (int, optional): min_elev_diff. Defaults to 200.
            min_horizonal_scale (int, optional): min_horizonal_scale. Defaults to 10000.
            vertical_scale (int, optional): vertical_scale. Defaults to 200.
            pos (int, optional): pos. Defaults to 4.
            neg (int, optional): neg. Defaults to 8.
            eps2 (float, optional): eps2. Defaults to 0.5.
            cmin (float, optional): cmin. Defaults to 0.9.
            cmax (float, optional): cmax. Defaults to 1.1.
            missing_elev_to_zero (bool, optional): Set missing elevtions to zero. Defaults to False.

        """
        self.num_min = int(num_min)
        self.num_max = int(num_max)
        self.inner_radius = float(inner_radius)
        self.outer_radius = float(outer_radius)
        self.num_iterations = int(num_iterations)
        self.num_min_prof = int(num_min_prof)
        self.min_elev_diff = float(min_elev_diff)
        self.min_horizonal_scale = float(min_horizonal_scale)
        self.vertical_scale = float(vertical_scale)
        self.def_pos = float(pos)
        self.def_neg = float(neg)
        self.def_eps2 = float(eps2)
        self.cmin = cmin
        self.cmax = cmax
        self.pos = []
        self.neg = []
        self.eps2 = []
        self.missing_elev_to_zero = missing_elev_to_zero
        QualityControl.__init__(self, "sct")

    def set_input(self, size, neg=None, pos=None, eps2=None):
        """Set input.

        Args:
            size (int): Observation set size.
            neg (float, optional): Minimum value. Defaults to None.
            pos (float, optional): Maximum value. Defaults to None.
            eps2 (float, optional): Epsilon 2. Defaults to None.

        """
        used_pos = self.def_pos
        if pos is not None:
            used_pos = pos

        used_neg = self.def_neg
        if neg is not None:
            used_neg = neg

        used_eps2 = self.def_eps2
        if eps2 is not None:
            used_eps2 = eps2

        logging.debug("pos: %s", used_pos)
        logging.debug("neg: %s", used_neg)
        logging.debug("eps2: %s", used_eps2)

        pos = []
        neg = []
        eps2 = []
        for __ in range(0, size):
            pos.append(used_pos)
            neg.append(used_neg)
            eps2.append(used_eps2)

        self.pos = self.pos + pos
        self.neg = self.neg + neg
        self.eps2 = self.eps2 + eps2

    def test(self, dataset, mask, code=105):
        """Do the test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 105.

        Raises:
            ModuleNotFoundError: titanlib was not loaded properly
            RuntimeError: Longitude is not defined!
            RuntimeError: Latitude is not defined!
            RuntimeError: Value is not defined!

        Returns:
            global_flags(list): Global flags.

        """
        if tit is None:
            raise ModuleNotFoundError("titanlib was not loaded properly")

        global_flags = dataset.flags
        lons = []
        lats = []
        elevs = []
        values = []

        old_mask = mask
        mask = []
        nmissing_elev = 0
        for i, old_mask_ind in enumerate(old_mask):
            if np.isnan(dataset.elevs[old_mask_ind]):
                if not self.missing_elev_to_zero:
                    ind = i - nmissing_elev
                    self.pos.pop(ind)
                    self.neg.pop(ind)
                    self.eps2.pop(ind)
                else:
                    mask.append(old_mask_ind)
                nmissing_elev = nmissing_elev + 1
            else:
                mask.append(old_mask_ind)

        for i, mask_ind in enumerate(mask):
            lons.append(dataset.lons[mask_ind])
            lats.append(dataset.lats[mask_ind])
            if np.isnan(dataset.elevs[mask_ind] and self.missing_elev_to_zero):
                elevs.append(0)
            else:
                elevs.append(dataset.elevs[mask_ind])
            values.append(dataset.values[mask_ind])

            # DEBUG
            if np.isnan(dataset.lons[mask_ind]):
                logging.error(i, "lon")
                raise RuntimeError("Longitude is not defined!")

            if np.isnan(dataset.lats[mask_ind]):
                logging.error(i, "lat")
                raise RuntimeError("Latitude is not defined!")

            if np.isnan(dataset.values[mask_ind]):
                logging.error(i, "value")
                raise RuntimeError("Value is not defined!")

        if nmissing_elev > 0:
            if self.missing_elev_to_zero:
                logging.info(
                    "Found %s / %s observations with undefined elevations which were "
                    "set to zero",
                    str(nmissing_elev),
                    str(len(old_mask)),
                )
            else:
                logging.info(
                    "Removed %s / %s obsevations with undefined elevations",
                    str(nmissing_elev),
                    str(len(old_mask)),
                )

        logging.info("Running sct")
        if len(values) > 0:
            lats = np.asarray(lats)
            lons = np.asarray(lons)
            elevs = np.asarray(elevs)
            values = np.asarray(values)
            points = tit.Points(lats, lons, elevs)
            answer = tit.sct(
                points,
                values,
                self.num_min,
                self.num_max,
                self.inner_radius,
                self.outer_radius,
                self.num_iterations,
                self.num_min_prof,
                self.min_elev_diff,
                self.min_horizonal_scale,
                self.vertical_scale,
                self.pos,
                self.neg,
                self.eps2,
            )

            flags = answer[0]
            sct = answer[1]
            rep = answer[2]
            for i, mask_ind in enumerate(mask):
                if int(global_flags[mask_ind]) == 0 and flags[i] == 1:
                    global_flags[mask_ind] = code

            dataset.normalize_ci(mask, self.cmin, self.cmax)

            for i, mask_ind in enumerate(mask):
                logging.debug(
                    "test=%s i=%s m_i=%s value(m_i)=%s sct(i)=%s rep(i)=%s flag(i)=%s "
                    "global_flag(m_i)=%s",
                    self.name,
                    i,
                    mask_ind,
                    dataset.values[mask_ind],
                    sct[i],
                    rep[i],
                    int(flags[i]),
                    int(global_flags[mask_ind]),
                )
        else:
            logging.info("No observations to run test on")

        return global_flags


class Buddy(QualityControl):
    """Buddy test."""

    def __init__(
        self,
        diff_elev_max=200000.0,
        adjust_for_elev_diff=True,
        distance_lim=1000000.0,
        priorities=1,
        buddies_min=1,
        thresholds=1.0,
        obs_to_check=1,
    ):
        """Construct buddy test.

        Args:
            diff_elev_max (_type_, optional): diff_elev_max. Defaults to 200000..
            adjust_for_elev_diff (bool, optional): adjust_for_elev_diff. Defaults to True.
            distance_lim (_type_, optional): distance_lim. Defaults to 1000000..
            priorities (int, optional): priorities. Defaults to 1.
            buddies_min (int, optional): buddies_min. Defaults to 1.
            thresholds (_type_, optional): thresholds. Defaults to 1..
            obs_to_check (int, optional): obs_to_check. Defaults to 1.

        """
        self.diff_elev_max = diff_elev_max
        self.adjust_for_elev_diff = adjust_for_elev_diff
        self.distance_lim = []
        self.priorities = []
        self.buddies_min = []
        self.thresholds = []
        self.obs_to_check = []
        self.def_distance_lim = distance_lim
        self.def_priorities = priorities
        self.def_buddies_min = buddies_min
        self.def_thresholds = thresholds
        self.def_obs_to_check = obs_to_check
        QualityControl.__init__(self, "buddy")

    def set_input(
        self,
        size,
        distance_lim=None,
        priorities=None,
        buddies_min=None,
        thresholds=None,
        obs_to_check=None,
    ):
        """Set input.

        Args:
            size (int): Observation set size.
            distance_lim (int, optional): distance_lim. Defaults to None.
            priorities (int, optional): priorities. Defaults to None.
            buddies_min (int, optional): buddies_min. Defaults to None.
            thresholds (float, optional): thresholds. Defaults to None.
            obs_to_check (int, optional): obs_to_check. Defaults to None.

        """
        used_distance_lim = self.def_distance_lim
        used_priorities = self.def_priorities
        used_buddies_min = self.def_buddies_min
        used_thresholds = self.def_thresholds
        used_obs_to_check = self.def_obs_to_check
        distance_lim = []
        priorities = []
        buddies_min = []
        thresholds = []
        obs_to_check = []

        if distance_lim is not None:
            used_distance_lim = distance_lim
        if priorities is not None:
            used_distance_lim = priorities
        if buddies_min is not None:
            used_distance_lim = buddies_min
        if thresholds is not None:
            used_distance_lim = thresholds
        if obs_to_check is not None:
            used_distance_lim = obs_to_check

        for __ in range(0, size):
            distance_lim.append(used_distance_lim)
            priorities.append(used_priorities)
            buddies_min.append(used_buddies_min)
            thresholds.append(used_thresholds)
            obs_to_check.append(used_obs_to_check)

        logging.debug("distance_lim: %s", used_distance_lim)
        logging.debug("  priorities: %s", used_priorities)
        logging.debug(" buddies_min: %s", used_buddies_min)
        logging.debug("  thresholds: %s", used_thresholds)
        logging.debug("obs_to_check: %s", used_obs_to_check)

        self.distance_lim = self.distance_lim + distance_lim
        self.priorities = self.priorities + priorities
        self.buddies_min = self.buddies_min + buddies_min
        self.thresholds = self.thresholds + thresholds
        self.obs_to_check = self.obs_to_check + obs_to_check

    def test(self, dataset, mask, code=104):
        """Do the test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 104.

        Raises:
            ModuleNotFoundError: titanlib was not loaded properly
            RuntimeError: Buddy check failed!

        Returns:
            global_flags(list): Global flags.

        """
        if tit is None:
            raise ModuleNotFoundError("titanlib was not loaded properly")

        global_flags = dataset.flags
        # Buddy does not work properly for dataset.
        # Also without data set the values must be set without subscripts

        lons = []
        lats = []
        elevs = []
        values = []
        for i in range(0, len(mask)):
            lons.append(dataset.lons[i])
            lats.append(dataset.lats[i])
            elevs.append(dataset.elevs[i])
            values.append(dataset.values[i])

        points = tit.Points(lats, lons, elevs)
        status, flags = tit.buddy_check(
            points,
            values,
            self.distance_lim,
            self.priorities,
            self.buddies_min,
            self.thresholds,
            self.diff_elev_max,
            self.adjust_for_elev_diff,
            self.obs_to_check,
        )
        if not status:
            raise RuntimeError("Buddy check failed!")

        for i, mask_ind in enumerate(mask):
            if global_flags[mask_ind] == 0 and flags[i] == 1:
                global_flags[mask_ind] = code

        for i, mask_ind in enumerate(mask):
            logging.debug(
                "test=%s i=%s m_i=%s value=%s flag(i)=%s global_flag(m_i)=%s",
                self.name,
                i,
                mask_ind,
                dataset.values[i],
                dataset.flags[i],
                global_flags[mask_ind],
            )

        return global_flags


class Climatology(QualityControl):
    """Climatology QC test."""

    def __init__(self, an_time, minval=None, maxval=None, offset=0):
        """Construct test.

        Args:
            an_time (datetime.datetime): Analysis time
            minval (float, optional): Minimum value. Defaults to None.
            maxval (float, optional): Maximum value. Defaults to None.
            offset (int, optional): Offset. Defaults to 0.

        """
        if isinstance(an_time, str):
            an_time = as_datetime(an_time)
        self.unixtime = int(an_time.strftime("%s"))
        self.def_min = minval
        self.def_max = maxval
        self.def_offset = offset
        self.offset = []
        self.minvals = []
        self.maxvals = []
        QualityControl.__init__(self, "climatology")

    def set_input(self, size, minval=None, maxval=None, offset=None):
        """Set input.

        Args:
            size (int): Observation set size.
            minval (float, optional): Minimum value. Defaults to None.
            maxval (float, optional): Maximum value. Defaults to None.
            offset (int, optional): Offset. Defaults to 0.

        Raises:
            RuntimeError: You must set min and max values!

        """
        used_min = self.def_min
        used_max = self.def_max
        used_offset = self.def_offset
        if minval is not None:
            used_min = minval
        if maxval is not None:
            used_max = maxval
        if offset is not None:
            used_offset = offset
        if used_min is None or used_max is None:
            raise RuntimeError("You must set min and max values!")

        minvals = []
        maxvals = []
        offset = []
        for __ in range(0, size):
            minvals.append(used_min)
            maxvals.append(used_max)
            offset.append(used_offset)
        self.minvals = self.minvals + minvals
        self.maxvals = self.maxvals + maxvals
        self.offset = self.offset + offset

    def test(self, dataset, mask, code=103):
        """Do the test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 103.

        Returns:
            global_flags(list): Global flags.

        """
        lons = []
        lats = []
        elevs = []
        values = []
        for obs, mask_ind in enumerate(mask):
            lons.append(dataset.lons[mask_ind])
            lats.append(dataset.lats[mask_ind])
            elevs.append(dataset.elevs[mask_ind])
            val = dataset.values[mask_ind] + self.offset[obs]
            values.append(val)

        points = tit.Points(lats, lons, elevs)
        flags = tit.range_check_climatology(
            points, values, self.unixtime, self.maxvals, self.minvals
        )

        global_flags = dataset.flags
        for i, mask_ind in enumerate(mask):
            if int(global_flags[mask_ind]) == 0 and int(flags[i]) == 1:
                global_flags[mask_ind] = code

        for i, mask_ind in enumerate(mask):
            logging.debug(
                "test=%s i=%s m_i=%s lon(i)=%s lat(i)=%s elev(i)=%s min_val(i)=%s "
                "value(i)=%s maxval(i)=%s value(m_i)=%s flag(i)=%s globalflag(m_i)=%s",
                self.name,
                i,
                mask_ind,
                lons[i],
                lats[i],
                elevs[i],
                self.minvals[i],
                values[i],
                self.maxvals[i],
                dataset.values[mask_ind],
                flags[i],
                global_flags[mask_ind],
            )
        return global_flags


class Redundancy(QualityControl):
    """Redundancy."""

    def __init__(self, an_time):
        """Construct test.

        Args:
            an_time (datetime.datetime): Analysis time.
        """
        self.an_time = an_time
        QualityControl.__init__(self, "redundancy")

    def set_input(self, __):
        """Set input."""

    def test(self, dataset, mask, code=115):
        """Do the test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 115.

        Returns:
            flags(list): Flags.

        """
        data = {}
        flags = dataset.flags
        for i, lon_val in enumerate(dataset.lons):
            if i in mask:
                lon1 = f"{float(lon_val):10.5f}"
                lat1 = f"{float(dataset.lats[i]):10.5f}"
                obstime1 = dataset.obstimes[i]
                pos = str(lon1) + ":" + str(lat1)

                if pos in data:
                    obstime = data[pos]["obstime"]
                    logging.debug(
                        "Found a redundant observation %s %s %s %s",
                        i,
                        pos,
                        dataset.stids[i],
                        obstime1,
                    )
                    # New best position in time. Flag the previous
                    if abs(self.an_time - obstime1) < abs(self.an_time - obstime):
                        logging.debug(
                            "Found a better redundant observation %s %s %s",
                            pos,
                            obstime1,
                            obstime,
                        )
                        ind = data[pos]["index"]
                        flags[ind] = code
                        data.update({pos: {"obstime": obstime, "index": i}})
                    else:
                        flags[i] = code
                else:
                    data.update({pos: {"obstime": obstime1, "index": i}})

        for i, mask_ind in enumerate(mask):
            logging.debug(
                "test=%s i=%s m_i=%s time(m_i)=%s value(m_i)=%s flags(i)=%s ",
                self.name,
                i,
                mask_ind,
                dataset.obstimes[mask_ind],
                dataset.values[mask_ind],
                flags[mask_ind],
            )

        return flags


class Blacklist(QualityControl):
    """Blacklist."""

    def __init__(self, blacklist):
        """Construct test.

        Args:
            blacklist (dict): Blacklist positions/stids

        Raises:
            RuntimeError: You must set blacklist as a dict
            RuntimeError: Blacklist must have the same length for both lons and lats

        """
        if blacklist is None or not isinstance(blacklist, dict):
            raise RuntimeError("You must set blacklist as a dict")

        blacklist_pos = {}
        blacklist_stid = {}
        if "lons" in blacklist and "lats" in blacklist:
            for i in range(0, len(blacklist["lons"])):

                if len(blacklist["lons"]) != len(blacklist["lats"]):
                    raise RuntimeError(
                        "Blacklist must have the same length for both lons and lats"
                    )

                lon = Observation.format_lon(float(blacklist["lons"][i]))
                lat = Observation.format_lat(float(blacklist["lats"][i]))
                pos = str(lon) + ":" + str(lat)
                blacklist_pos.update({pos: 1})

        if "stids" in blacklist:
            for i in range(0, len(blacklist["stids"])):
                stid = str(blacklist["stids"][i])

                if stid != "NA":
                    blacklist_stid.update({str(stid): 1})

        self.blacklist_pos = blacklist_pos
        self.blacklist_stid = blacklist_stid
        QualityControl.__init__(self, "blacklist")

    def set_input(self, __):
        """Set input."""

    def test(self, dataset, mask, code=100):
        """Do the test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 100.

        Returns:
            flags(list): Flags.

        """
        flags = dataset.flags
        for i, lon_val in enumerate(dataset.lons):
            if i in mask:

                lon = Observation.format_lon(lon_val)
                lat = Observation.format_lat(dataset.lats[i])
                stid = dataset.stids[i]
                pos = lon + ":" + lat

                logging.debug("lon=%s lat=%s stid=%s pos=%s", lon, lat, stid, pos)
                if pos in self.blacklist_pos:
                    logging.debug("Found blacklisted position: %s", pos)
                    flags[i] = code
                if str(stid) in self.blacklist_stid:
                    logging.debug("Found blacklisted stid: %s", str(stid))
                    flags[i] = code

        for i, mask_ind in enumerate(mask):
            logging.debug(
                "test=%s i=%s m_i=%s lon(m_i)=%s lat(m_i)=%s stid(m_i)=%s flag(m_i)=%s",
                self.name,
                i,
                mask_ind,
                dataset.lons[mask_ind],
                dataset.lats[mask_ind],
                dataset.stids[mask_ind],
                flags[mask_ind],
            )

        return flags


class DomainCheck(QualityControl):
    """Domain check."""

    def __init__(self, domain_geo, max_distance=5000):
        """Construct test.

        Args:
            domain_geo (surfex.Geo): Surfex geometry
            max_distance (int, optional): Maximum distance to grid border. Defaults to 5000.

        Raises:
             RuntimeError: Domain geo was not set!

        """
        if domain_geo is None:
            raise RuntimeError("Domain geo was not set!")

        self.lons = domain_geo.lons
        self.lats = domain_geo.lats
        self.max_distance = max_distance
        QualityControl.__init__(self, "domain")

    def set_input(self, size, **kwargs):
        """Set input."""

    def test(self, dataset, mask, code=199):
        """Do the test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 199.

        Returns:
            flags(list): Flags.

        """
        flags = dataset.flags
        in_grid = inside_grid(
            self.lons, self.lats, dataset.lons, dataset.lats, distance=self.max_distance
        )
        # TODO vectorize
        for __, mask_ind in enumerate(mask):
            if not in_grid[mask_ind]:
                flags[mask_ind] = code

        for i, mask_ind in enumerate(mask):
            logging.debug(
                "test=%s i=%s m_i=%s lon(m_i)=%s lat(m_i)=%s stid(m_i)=%s flag(m_i)=%s",
                self.name,
                i,
                mask_ind,
                dataset.lons[mask_ind],
                dataset.lats[mask_ind],
                dataset.stids[mask_ind],
                flags[mask_ind],
            )
        return flags


class NoMeta(QualityControl):
    """Check for missing meta data."""

    def __init__(self):
        """Construct test."""
        QualityControl.__init__(self, "nometa")

    def set_input(self, size, **kwargs):
        """Set input."""

    def test(self, dataset, mask, code=101):
        """Do the test.

        Args:
            dataset (QCDataSet): The data set to perform the test on.
            mask (list): Active data.
            code (int, optional): Code to use for flagging. Defaults to 101.

        Returns:
            flags(list): Flags.

        """
        flags = dataset.flags
        for mask_ind in mask:
            if np.isnan(dataset.elevs[mask_ind]):
                flags[mask_ind] = code

        for i, mask_ind in enumerate(mask):
            logging.debug(
                "test=%s i=%s m_i=%s lon(m_i)=%s lat(m_i)=%s stid(m_i)=%s flag(m_i)=%s",
                self.name,
                i,
                mask_ind,
                dataset.lons[mask_ind],
                dataset.lats[mask_ind],
                dataset.stids[mask_ind],
                flags[mask_ind],
            )

        return flags


def define_quality_control(test_list, settings, an_time, domain_geo=None, blacklist=None):
    """Define different QC test from a dict.

    Parameters:
        test_list(list): List of tests
        settings(dict): Test settings
        an_time(datetime.datetime): Analysis time
        domain_geo(surfex.Geo): Geo object
        blacklist(dict): Optional blacklist. Needd for blacklist test

    Raises:
        NotImplementedError: Test not implemented
        RuntimeError: You must set the name of fg file and variable
        RuntimeError: You must set the name of fraction file and variable
        RuntimeError: You must set both fraction_field and fraction_geo
        RuntimeError: You must set both fg_field and fg_geo

    Returns:
        tests(list): List of QualityControl objects

    """
    tests = []
    for qct in test_list:
        logging.info("Set up test: %s", qct)
        kwargs = {}
        test_options = None
        if qct in settings:
            test_options = settings[qct]

        if qct.lower() == "plausibility":
            if test_options is not None:
                opts = ["minval", "maxval"]
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})
            tests.append(Plausibility(**kwargs))

        elif qct.lower() == "firstguess":
            fg_file = None
            fg_var = None
            fg_field = None
            fg_geo = None
            if test_options is not None:
                if "fg_file" in test_options:
                    fg_file = test_options["fg_file"]
                if "fg_var" in test_options:
                    fg_var = test_options["fg_var"]
                if "fg_field" in test_options:
                    fg_field = test_options["fg_field"]
                if "fg_geo" in test_options:
                    fg_geo = test_options["fg_geo"]
                opts = ["negdiff", "posdiff", "max_distance", "operator"]
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})
            if fg_geo is None and fg_field is None:
                if fg_file is None or fg_var is None:
                    raise RuntimeError("You must set the name of fg file and variable")
                fg_geo, __, fg_field, __, __ = read_first_guess_netcdf_file(
                    fg_file, fg_var
                )
            else:
                if fg_geo is None or fg_field is None:
                    raise RuntimeError("You must set both fg_field and fg_geo")
            tests.append(FirstGuess(fg_geo, fg_field, **kwargs))

        elif qct.lower() == "fraction":
            kwargs.update({"minval": 0.99, "maxval": 1.01})
            fraction_var = None
            fraction_file = None
            fraction_field = None
            fraction_geo = None
            if test_options is not None:
                if "fraction_file" in test_options:
                    fraction_file = test_options["fraction_file"]
                if "fraction_var" in test_options:
                    fraction_var = test_options["fraction_var"]
                if "fraction_field" in test_options:
                    fraction_field = test_options["fraction_field"]
                if "fraction_geo" in test_options:
                    fraction_geo = test_options["fraction_geo"]
                opts = ["minval", "maxval", "max_distance", "operator"]
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})

            if fraction_geo is None and fraction_field is None:
                if fraction_var is None or fraction_file is None:
                    raise RuntimeError(
                        "You must set the name of fraction file and variable"
                    )

                fraction_geo, __, fraction_field, __, __ = read_first_guess_netcdf_file(
                    fraction_file, fraction_var
                )
            else:
                if fraction_field is None or fraction_geo is None:
                    raise RuntimeError(
                        "You must set both fraction_field and fraction_geo"
                    )
            tests.append(Fraction(fraction_geo, fraction_field, **kwargs))

        elif qct.lower() == "buddy":
            if test_options is not None:
                opts = [
                    "diff_elev_max",
                    "adjust_for_elev_diff",
                    "distance_lim",
                    "priorities",
                    "buddies_min",
                    "thresholds",
                    "obs_to_check",
                ]
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})
            tests.append(Buddy(**kwargs))

        elif qct.lower() == "climatology":

            if test_options is not None:
                opts = ["minval", "maxval", "offset"]
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})
            tests.append(Climatology(an_time, **kwargs))

        elif qct.lower() == "sct":

            if test_options is not None:
                opts = [
                    "num_min",
                    "num_max",
                    "inner_radius",
                    "outer_radius",
                    "num_iterations",
                    "num_min_prof",
                    "min_elev_diff",
                    "min_horizontal_scale",
                    "pos",
                    "neg",
                    "eps2",
                    "cmin",
                    "cmax",
                ]
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})
            tests.append(Sct(**kwargs))

        elif qct.lower() == "redundancy":
            if test_options is not None:
                opts = []
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})
            tests.append(Redundancy(an_time, **kwargs))

        elif qct.lower() == "blacklist":
            if test_options is not None:
                opts = []
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})
            tests.append(Blacklist(blacklist, **kwargs))

        elif qct.lower() == "domain":
            if test_options is not None:
                opts = ["max_distance"]
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})
            tests.append(DomainCheck(domain_geo, **kwargs))

        elif qct.lower() == "nometa":
            if test_options is not None:
                opts = []
                for opt in opts:
                    if opt in test_options:
                        kwargs.update({opt: test_options[opt]})
            tests.append(NoMeta(**kwargs))

        else:
            raise NotImplementedError(f"Test {qct} is not implemented")

    return tests


class QCDataSet(object):
    """QC data set."""

    def __init__(
        self,
        analysis_time,
        observations,
        flags,
        cis,
        lafs,
        providers,
        passed_tests=None,
        fg_dep=None,
        an_dep=None,
        remove_invalid_elevs=False,
    ):
        """Construct QC data set.

        Args:
            analysis_time (datetime.datetime): Analysis time
            observations (list): Observation objects.
            flags (list): Flags
            cis (list): Corep
            lafs (list): Land area fraction
            providers (list): Providers.
            passed_tests (list, optional): Tests which have been passed. Defaults to None.
            fg_dep (list, optional): First guess departures. Defaults to None.
            an_dep (list, optional): Analysis depatures. Defaults to None.
            remove_invalid_elevs (bool, optional): Remove invalid elevations. Defaults to False.

        """
        self.analysis_time = analysis_time
        self.index_pos = {}
        self.index_stid = {}
        obstimes = []
        lons = []
        lats = []
        elevs = []
        stids = []
        values = []
        varnames = []
        for i, observation in enumerate(observations):
            obstimes.append(observation.obstime)
            lons.append(observation.lon)
            lats.append(observation.lat)
            stids.append(observation.stid)
            elevs.append(observation.elev)
            values.append(observation.value)
            varnames.append(observation.varname)

            lon = f"{float(observation.lon):10.5f}"
            lat = f"{float(observation.lat):10.5f}"
            stid = str(observation.stid)

            pos = lon + ":" + lat
            self.index_pos.update({pos: i})
            if stid != "NA":
                self.index_stid.update({stid: i})

        self.obstimes = obstimes
        self.lons = lons
        self.lats = lats
        self.elevs = elevs
        self.stids = stids
        self.values = values
        self.varnames = varnames

        self.flags = flags

        metadata = 0
        for i in range(0, len(flags)):
            if remove_invalid_elevs and np.isnan(elevs[i]):
                self.flags[i] = 101
                metadata = metadata + 1
        self.metadata = metadata
        self.cis = cis
        self.lafs = lafs
        self.providers = providers
        if passed_tests is None:
            passed_tests = []
            for __ in range(0, len(observations)):
                passed_tests.append([])
            self.passed_tests = passed_tests
        else:
            self.passed_tests = passed_tests
        if fg_dep is None:
            fg_dep = []
            for __ in range(0, len(observations)):
                fg_dep.append(np.nan)
        self.fg_dep = fg_dep
        if an_dep is None:
            an_dep = []
            for __ in range(0, len(observations)):
                an_dep.append(np.nan)
        self.an_dep = an_dep

    def get_stid_index(self, stid):
        """Get station ID index.

        Args:
            stid (str): Station ID

        Returns:
            int: Index to station ID.
        """
        stid = str(stid)
        if stid in self.index_stid:
            return self.index_stid[stid]
        return None

    def get_pos_index(self, lon, lat):
        """Get position index in dict.

        Args:
            lon (float): Longitude
            lat (float): Latitude

        Returns:
            int: Index position if found, else None.
        """
        lon = f"{float(lon):10.5f}"
        lat = f"{float(lat):10.5f}"
        pos = lon + ":" + lat
        if pos in self.index_pos:
            return self.index_pos[pos]
        return None

    @abc.abstractmethod
    def perform_tests(self):
        """Perform the test.

        Raises:
            NotImplementedError: Must be implemented by child class.
        """
        raise NotImplementedError("You must implement this method")

    def write_output(self, filename, indent=None):
        """Dump QC data to a json file.

        Args:
            filename (str): Filename
            indent (int, optional): Indentation in file. Defaults to None.
        """
        data = {}
        for i, lon_val in enumerate(self.lons):
            data.update(
                {
                    i: {
                        "varname": self.varnames[i],
                        "obstime": as_datetime_string(self.obstimes[i]),
                        "lon": lon_val,
                        "lat": self.lats[i],
                        "stid": self.stids[i],
                        "elev": self.elevs[i],
                        "value": self.values[i],
                        "flag": self.flags[i],
                        "ci": self.cis[i],
                        "laf": self.lafs[i],
                        "provider": self.providers[i],
                        "fg_dep": self.fg_dep[i],
                        "an_dep": self.an_dep[i],
                        "passed_tests": self.passed_tests[i],
                    }
                }
            )
        json.dump(data, open(filename, mode="w", encoding="utf-8"), indent=indent)

    def normalize_ci(self, mask, cmin, cmax):
        """Normalize the cross correlation.

        Args:
            mask (list): Mask of active observations
            cmin (float): Minimum corelation
            cmax (float): Maximum correaltion.

        """
        nsize = len(mask)
        corep = []
        for i in range(0, nsize):
            corep.append(self.cis[mask[i]])
        corep = np.asarray(corep)
        if nsize > 0:

            def ecdf(xxx):
                xxx = np.sort(xxx)
                nnn = len(xxx)

                def _ecdf(vvv):
                    # side='right' because we want Pr(x <= v)
                    return (np.searchsorted(xxx, vvv, side="right") + 1) / nnn

                return _ecdf

            qmn = 0.25
            qmx = 0.75
            qav = 0.5

            acorep = abs(corep)
            fun = ecdf(acorep)
            qcorep = fun(acorep)

            qcorep[qcorep < qmn] = qmn
            qcorep[qcorep > qmx] = qmx

            corep_max = cmax * np.ones(nsize)
            corep_mean = 1 * np.ones(nsize)
            corep_min = cmin * np.ones(nsize)

            lav = np.where(qcorep <= qav)[0].tolist()
            if len(lav) > 0:
                corep1 = (qcorep[lav] - qmn) / (qav - qmn)
                corep[lav] = corep_min[lav] + (corep_mean[lav] - corep_min[lav]) * corep1

            gav = np.where(qcorep > qav)[0].tolist()
            if len(gav) > 0:
                corep1 = (qcorep[gav] - qav) / (qmx - qav)
                corep[gav] = corep_mean[gav] + (corep_max[gav] - corep_mean[gav]) * corep1

        for i in range(0, nsize):
            self.cis[mask[i]] = qcorep[i]


class TitanDataSet(QCDataSet):
    """Titan QC data set."""

    def __init__(
        self, var, settings, tests, datasources, an_time, corep=1, test_flags=None
    ):
        """Titan Data set.

        Args:
            var (str): Variable name.
            settings (dict): Titan test/configuration settings
            tests (list): Tests to perform in order.
            datasources (list): List of observations sets.
            an_time (datetime.datetime): Analysis time
            corep (int, optional): Correlations between observations. Defaults to 1.
            test_flags (dict, optional): Dictionary to set custom test flags. Defaults to None.

        Raises:
            ModuleNotFoundError: itanlib was not loaded properly

        """
        if tit is None:
            raise ModuleNotFoundError("titanlib was not loaded properly")

        self.var = var
        self.tests = tests
        self.settings = settings
        self.test_flags = test_flags
        obstimes = []
        lons = []
        lats = []
        stids = []
        elevs = []
        values = []
        varnames = []
        providers = []
        passed_tests = []
        self.datasources = datasources

        # Get global data
        for obs_set in self.datasources:
            (
                lobstimes,
                llons,
                llats,
                lstids,
                lelevs,
                lvalues,
                lvarnames,
            ) = obs_set.get_obs()
            obstimes = obstimes + lobstimes
            lons = lons + llons
            lats = lats + llats
            stids = stids + lstids
            elevs = elevs + lelevs
            values = values + lvalues
            varnames = varnames + lvarnames
            for __ in range(0, len(llons)):
                providers.append(obs_set.label)

        for __ in range(0, len(lons)):
            passed_tests.append([])
        flags = np.zeros(len(lons))
        cis = np.ones(len(lons)) * corep
        lafs = np.ones(len(lons))

        observations = []
        for i, lon_val in enumerate(lons):
            observations.append(
                Observation(
                    obstimes[i],
                    lon_val,
                    lats[i],
                    values[i],
                    elev=elevs[i],
                    stid=stids[i],
                    varname=varnames[i],
                )
            )
        points = tit.Points(lats, lons, elevs)
        self.titan_dataset = tit.Dataset(points, values)
        if passed_tests is None:
            passed_tests = []
            for __ in range(0, len(lons)):
                passed_tests.append([])
            self.passed_tests = passed_tests
        else:
            self.passed_tests = passed_tests

        QCDataSet.__init__(
            self,
            an_time,
            observations,
            flags,
            cis,
            lafs,
            providers,
            passed_tests=None,
            remove_invalid_elevs=False,
        )

    def perform_tests(self):
        """Perform the tests."""
        summary = {}
        for test in self.tests:
            print("Test: ", test.name)
            mask = []
            findex = 0
            for obs_set in self.datasources:

                if obs_set.label == "":
                    raise RuntimeError(
                        "Observations set for quality control are "
                        "assumed to have a label"
                    )

                print("obs_set", obs_set.label)
                size = obs_set.size

                do_test = False
                if "do_test" in self.settings:
                    do_test = self.settings["do_test"]
                if test.name in self.settings:
                    if "do_test" in self.settings[test.name]:
                        do_test = self.settings[test.name]["do_test"]

                test_settings = {"do_test": do_test}
                if "sets" in self.settings:
                    if obs_set.label in self.settings["sets"]:
                        if "tests" in self.settings["sets"][obs_set.label]:
                            if test.name in self.settings["sets"][obs_set.label]["tests"]:
                                test_settings.update(
                                    self.settings["sets"][obs_set.label]["tests"][
                                        test.name
                                    ]
                                )

                do_test = test_settings["do_test"]
                if do_test:
                    del test_settings["do_test"]
                    logging.debug("findex %s size %s", findex, size)
                    lmask = np.where(np.asarray(self.flags[findex : findex + size]) == 0)[
                        0
                    ].tolist()

                    for lmask_ind in lmask:
                        lmask_ind = lmask_ind + findex
                    mask = mask + lmask

                    # Set input for this set
                    logging.info(
                        "Test %s size=%s settings=%s", test.name, len(mask), test_settings
                    )
                    test.set_input(len(lmask), **test_settings)

                else:
                    logging.info(
                        "Test %s is de-ativated for this data source %s",
                        test.name,
                        obs_set.label,
                    )

                findex = findex + size

            # Tests on active observations
            ok_obs = 0
            bad = 0
            outside = 0
            if len(mask) > 0:
                kwargs = {}
                if self.test_flags is not None:
                    if test.name in self.test_flags:
                        kwargs.update({"code": self.test_flags[test.name]})

                self.flags = test.test(self, mask, **kwargs)
                for mask_ind in mask:
                    if self.flags[mask_ind] == 0:
                        self.passed_tests[mask_ind].append(test.name)
                        ok_obs = ok_obs + 1
                    else:
                        self.titan_dataset.flags[mask_ind] = 1
                        bad = bad + 1
                        if self.flags[mask_ind] == 199:
                            outside = outside + 1

            summary.update(
                {
                    test.name: {
                        "tested": len(mask),
                        "ok": ok_obs,
                        "bad": bad,
                        "outside": outside,
                    }
                }
            )

        kept = 0
        flagged = 0
        for flag in self.flags:
            if flag == 0:
                kept = kept + 1
            else:
                flagged = flagged + 1

        # Print summary
        logging.info("\n")
        logging.info("Total number of observations: %s", len(self.flags))
        logging.info("                        Kept: %s", kept)
        logging.info(
            "                     Flagged: %s (bad metadata: %s)", flagged, self.metadata
        )
        logging.info("\n")
        for test in self.tests:
            outside = ""
            if summary[test.name]["outside"] > 0:
                outside = (
                    " (" + str(summary[test.name]["outside"]) + " exceeding max distance)"
                )
            logging.info("Test: %s", test.name)
            logging.info("  tested: %s", summary[test.name]["tested"])
            logging.info("      ok: %s", summary[test.name]["ok"])
            logging.info("     bad: %s %s", summary[test.name]["bad"], outside)
            logging.info("\n")


class Departure(object):
    """Departure. Difference between an observation and a value."""

    def __init__(self, operator, geo, dataset, grid_values, mode, max_distance=5000):
        """Construct a departure object.

        Args:
            operator (str): Interpolation operator.
            geo (surfex.Geo): Surfex geometry.
            dataset (QCDataSet): QC data set.
            grid_values (np.darray): Values in the grid.
            mode (str): What kind of departure (analysis/first_guess)
            max_distance (int, optional): Max allowed deviation in meters from grid borders.
                                          Defaults to 5000.

        Raises:
            NotImplementedError: Mode not implemented

        """
        self.obs_operator = ObsOperator(
            operator, geo, dataset, grid_values, max_distance=max_distance
        )
        obs_values = self.obs_operator.get_obs_value()

        values = []
        departures = []
        for i, obs_value in enumerate(obs_values):
            if mode == "analysis":
                if int(dataset.flags[i]) == 0:
                    values.append(obs_value)
                else:
                    values.append(np.nan)
            elif mode == "first_guess":
                values.append(obs_value)
            else:
                raise NotImplementedError(mode)
            if np.isnan(values[i]):
                departures.append(np.nan)
            else:
                departures.append(dataset.values[i] - values[i])

        self.departures = departures
        self.obs_values = values

    def get_departure(self, pos=None):
        """Get departure.

        Args:
            pos (int, optional): Position. Defaults to None.

        Returns:
            list: Departures
        """
        if pos is None:
            return self.departures
        return None

    def get_values(self, pos=None):
        """Get the obs values.

        Args:
            pos (int, optional): Position. Defaults to None.

        Returns:
            float: Observation equivalent of the input field.

        """
        return self.obs_operator.get_obs_value(pos=pos)


def dataset_from_file(
    an_time, filename, qc_flag=None, skip_flags=None, fg_dep=None, an_dep=None
):
    """Get a QCDataSet from a json file.

    Args:
        an_time (datetime.datetime): Analysis time.
        filename (str): List of filenames.
        qc_flag (int, optional): QC code to merge. Defaults to None.
        skip_flags (list, optional): List of QC flags to skip. Defaults to None.
        fg_dep (dict, optional): First guess departures. Defaults to None.
        an_dep (dict, optional): Analysis departures. Defaults to None.

    Returns:
        QCDataSet: QCDataSet

    """
    data = json.load(open(filename, mode="r", encoding="utf-8"))
    return dataset_from_json(
        an_time,
        data,
        qc_flag=qc_flag,
        skip_flags=skip_flags,
        fg_dep=fg_dep,
        an_dep=an_dep,
    )


def dataset_from_json(
    an_time, data, qc_flag=None, skip_flags=None, fg_dep=None, an_dep=None
):
    """Create a QCDataSet data set read from a json file.

    Args:
        an_time (datetime.datetime): Analysis time.
        data (dict): data read from the file
        qc_flag (int, optional): QC code to merge. Defaults to None.
        skip_flags (list, optional): List of QC flags to skip. Defaults to None.
        fg_dep (dict, optional): First guess departures. Defaults to None.
        an_dep (dict, optional): Analysis departures. Defaults to None.

    Returns:
        QCDataSet: QCDataSet
    """
    observations = []
    providers = []
    flags = []
    cis = []
    lafs = []
    fg_deps = []
    an_deps = []
    passed_tests = []
    for i in data:
        add = False
        if qc_flag is not None:
            if data[i]["flag"] == qc_flag:
                add = True
        else:
            add = True

        if skip_flags is not None:
            for sfl in skip_flags:
                if int(data[i]["flag"]) == int(sfl):
                    add = False

        if add:
            obstime = as_datetime(data[i]["obstime"])
            lon = data[i]["lon"]
            lat = data[i]["lat"]
            stid = data[i]["stid"]
            elev = data[i]["elev"]
            value = data[i]["value"]
            observations.append(
                Observation(obstime, lon, lat, value, stid=stid, elev=elev)
            )
            if "provider" in data[i]:
                providers.append(data[i]["provider"])
            else:
                providers.append("NA")
            flag = data[i]["flag"]
            flags.append(flag)
            cis.append(data[i]["ci"])
            lafs.append(data[i]["laf"])
            if fg_dep is not None:
                fg_deps.append(fg_dep[int(i)])
            else:
                if "fg_dep" in data[i]:
                    fg_deps.append(data[i]["fg_dep"])
                else:
                    fg_deps.append(np.nan)
            if an_dep is not None:
                an_deps.append(an_dep[int(i)])
            else:
                if "an_dep" in data[i]:
                    an_deps.append(data[i]["an_dep"])
                else:
                    an_deps.append(np.nan)
            if "passed_tests" in data[i]:
                passed_tests.append(data[i]["passed_tests"])

    if len(passed_tests) == 0:
        passed_tests = None

    return QCDataSet(
        an_time,
        observations,
        flags,
        cis,
        lafs,
        providers,
        passed_tests=passed_tests,
        fg_dep=fg_deps,
        an_dep=an_deps,
    )


def merge_json_qc_data_sets(an_time, filenames, qc_flag=None, skip_flags=None):
    """Merge QC data sets from json files.

    Args:
        an_time (datetime.datetime): Analysis time.
        filenames (list): List of filenames.
        qc_flag (int, optional): QC code to merge. Defaults to None.
        skip_flags (list, optional): List of QC flags to skip. Defaults to None.

    Returns:
        QCDataSet: QCDataSet
    """
    ind = 0
    index_pos = {}
    data = {}
    for filename in filenames:

        if os.path.exists(filename):
            with open(filename, mode="r", encoding="utf-8") as file_handler:
                data1 = json.load(file_handler)
            for dd1 in data1:
                lon1 = data1[dd1]["lon"]
                lat1 = data1[dd1]["lat"]
                pos1 = f"{Observation.format_lon(lon1)}:{Observation.format_lat(lat1)}"
                if pos1 not in index_pos:
                    index_pos.update({pos1: ind})
                    data.update({str(ind): data1[dd1]})
                    ind = ind + 1
        else:
            logging.warning("File name does not exist: %s", filename)

    logging.info("Merged %s observations", str(ind))
    return dataset_from_json(an_time, data, qc_flag=qc_flag, skip_flags=skip_flags)



"""Time series."""
import json
import logging

from .datetime_utils import as_timedelta
from .obs import Observation


class TimeSeries(object):
    """Time series."""

    def __init__(self, times, values, lons, lats, stids, stids_file=None, varname="NA"):
        """Construct time series.

        Args:
            times (_type_): _description_
            values (_type_): _description_
            lons (_type_): _description_
            lats (_type_): _description_
            stids (_type_): _description_
            stids_file (_type_, optional): _description_. Defaults to None.
            varname (str, optional): _description_. Defaults to "NA".

        """
        self.times = times
        self.values = values
        self.lons = lons
        self.lats = lats
        self.stids = stids
        self.varname = varname
        self.index_pos = {}

        for i, lon in enumerate(self.lons):
            lat = self.lats[i]
            pos = Observation.format_lon(lon) + ":" + Observation.format_lat(lat)
            self.index_pos.update({pos: i})

    def write_json(self, filename, indent=None):
        """Write json file.

        Args:
            filename (_type_): _description_
            indent (_type_, optional): _description_. Defaults to None.
        """
        data = {}
        for i, time_val in enumerate(self.times):
            data.update(
                {time_val.strftime("%Y%m%d%H%M%S"): {"values": self.values[i].tolist()}}
            )
        data = {
            "lons": self.lons.tolist(),
            "lats": self.lats.tolist(),
            "stids": self.stids,
            "varname": self.varname,
            "data": data,
        }
        json.dump(data, open(filename, "w", encoding="utf-8"), indent=indent)


class TimeSeriesFromConverter(TimeSeries):
    """Time-Series from a converter."""

    def __init__(
        self, var, geo, converter, start, end, interval=3600, cache=None, stids_file=None
    ):
        """Construct.

        Args:
            var (_type_): _description_
            geo (_type_): _description_
            converter (Converter): _description_
            start (_type_): _description_
            end (_type_): _description_
            interval (int, optional): _description_. Defaults to 3600.
            cache (_type_, optional): _description_. Defaults to None.
            stids_file (_type_, optional): _description_. Defaults to None.

        """
        times = []
        values = []
        # Loop output time steps
        this_time = start
        while this_time <= end:
            # Write for each time step
            logging.info(
                "Creating time series for: %s time_step: %s",
                this_time.strftime("%Y%m%d%H"),
                str(this_time),
            )
            values.append(converter.read_time_step(geo, this_time, cache))
            times.append(this_time)

            this_time = this_time + as_timedelta(seconds=interval)
            if cache is not None:
                cache.clean_fields(this_time)

        if stids_file is not None:
            stids = Observation.get_stid_from_stationlist(
                stids_file, geo.lonlist, geo.latlist
            )
        else:
            stids = ["NA"] * geo.nlons

        TimeSeries.__init__(
            self,
            times,
            values,
            geo.lonlist,
            geo.latlist,
            stids,
            stids_file=stids_file,
            varname=var,
        )



"""Geometry."""
import json
import logging
import math
import os
from abc import ABC, abstractmethod

import numpy as np
import pyproj

try:
    from osgeo import ogr  # type: ignore
except Exception:
    ogr = None


from .namelist_legacy import BaseNamelist


class Geo(object):
    """Geometry."""

    def __init__(self, lons, lats):
        """Construct geometry.

        Args:
            lons (np.ndarray): Longitudes
            lats (np.ndarray): Latitudes

        Raises:
            Exception: _description_

        """
        can_interpolate = False
        if not isinstance(lons, np.ndarray) or not isinstance(lats, np.ndarray):
            raise Exception("Longitudes and latitudes must be numpy nd arrays")

        ndims = len(lons.shape)
        self.ndims = ndims
        if len(lons.shape) != len(lats.shape):
            raise Exception("Mismatch in lat and lon")
        if len(lons.shape) > 1 and len(lats.shape) > 1:
            can_interpolate = True

        logging.debug("ndims=%s, can_interpolate=%s", ndims, can_interpolate)
        nlons = lons.shape[0]
        nlats = lats.shape[0]
        if ndims > 1:
            nlats = lats.shape[1]

        if ndims > 1:
            npoints = nlons * nlats
        else:
            npoints = nlats
        self.npoints = npoints
        self.nlons = nlons
        self.nlats = nlats
        self.lonrange = [np.min(lons), np.max(lons)]
        self.latrange = [np.min(lats), np.max(lats)]
        self.lons = lons
        self.lonlist = lons.flatten()
        self.lats = lats
        self.latlist = lats.flatten()
        logging.debug("lonlist=%s", self.lonlist)
        logging.debug("latlist=%s", self.latlist)
        logging.debug("nlons=%s nlats=%s", nlons, nlats)
        logging.debug("lons=%s shape=%s", lons, lons.shape)
        logging.debug("lats=%s shape=%s", lats, lats.shape)
        self.can_interpolate = can_interpolate

    def identifier(self):
        """Create identifier."""
        f_lon = ""
        l_lon = ""
        if self.lonrange is not None:
            f_lon = str(round(float(self.lonrange[0]), 2))
            l_lon = str(round(float(self.lonrange[-1]), 2))
        f_lat = ""
        l_lat = ""
        if self.latrange is not None:
            f_lat = str(round(float(self.latrange[0]), 2))
            l_lat = str(round(float(self.latrange[-1]), 2))

        tag = (
            ":"
            + str(self.npoints)
            + ":"
            + str(self.nlons)
            + ":"
            + str(self.nlats)
            + ":"
            + f_lon
            + ":"
            + l_lon
            + ":"
            + f_lat
            + ":"
            + l_lat
            + ":"
        )
        tag = tag.replace(" ", "")
        logging.debug("TAG: %s", tag)
        return tag

    def is_identical(self, geo_to_check):
        """Check if geometries are identical.

        Args:
            geo_to_check (surfex.Geo): Geo to check.

        Returns:
            bool: True if identical.

        """
        if self.identifier() == geo_to_check.identifier():
            logging.debug("Geometries are identical")
            return True
        else:
            return False

    def write_proj_info(self):
        """Write proj info.

        Returns:
            None: Do nothing for now.
        """
        return None


class SurfexGeo(ABC, Geo):
    """Abstract surfex geometry class.

    Args:
        ABC (_type_): _description_
        Geo (_type_): _description_
    """

    def __init__(self, proj, lons, lats):
        """Construct a surfex geometry.

        Args:
            proj (str): Proj string
            lons (np.ndarray): Lons
            lats (np.ndarray): Lats

        """
        self.mask = None
        self.proj = proj
        Geo.__init__(self, lons, lats)

    @abstractmethod
    def update_namelist(self, nml):
        """Update namelist.

        Args:
            nml (_type_): _description_

        Returns:
            _type_: _description_
        """
        raise NotImplementedError

    @abstractmethod
    def subset(self, geo):
        """Find subset of geo.

        Args:
            geo (surfex.Geo): Geometry to check.
        """
        raise NotImplementedError


class ConfProj(SurfexGeo):
    """Conf proj."""

    def __init__(self, from_json):
        """Construct conf proj geo.

        Args:
            from_json (dict): Domain definition

        Raises:
            KeyError: Missing keys1
            KeyError: Missing keys2
            KeyError: Missing keys3
            KeyError: Missing keys4

        """
        self.cgrid = "CONF PROJ"
        self.json = from_json
        domain_dict = BaseNamelist.lower_case_namelist_dict(from_json)

        logging.debug("from_json: %s", from_json)
        self.ilone = None
        self.ilate = None
        self.xtrunc = None
        if "nam_conf_proj_grid" in domain_dict:
            if (
                "nimax"
                and "njmax"
                and "xloncen"
                and "xlatcen"
                and "xdx"
                and "xdy" in domain_dict["nam_conf_proj_grid"]
            ):
                self.nimax = domain_dict["nam_conf_proj_grid"]["nimax"]
                self.njmax = domain_dict["nam_conf_proj_grid"]["njmax"]
                self.xloncen = domain_dict["nam_conf_proj_grid"]["xloncen"]
                self.xlatcen = domain_dict["nam_conf_proj_grid"]["xlatcen"]
                self.xdx = domain_dict["nam_conf_proj_grid"]["xdx"]
                self.xdy = domain_dict["nam_conf_proj_grid"]["xdy"]
                if "ilone" in domain_dict["nam_conf_proj_grid"]:
                    self.ilone = domain_dict["nam_conf_proj_grid"]["ilone"]
                if "ilate" in domain_dict["nam_conf_proj_grid"]:
                    self.ilate = domain_dict["nam_conf_proj_grid"]["ilate"]
                if "xtrunc" in domain_dict["nam_conf_proj_grid"]:
                    self.xtrunc = domain_dict["nam_conf_proj_grid"]["xtrunc"]
            else:
                raise KeyError("Missing keys1")
        else:
            raise KeyError("Missing key2")

        if "nam_conf_proj" in domain_dict:
            if "xlon0" and "xlat0" in domain_dict["nam_conf_proj"]:
                self.xlon0 = domain_dict["nam_conf_proj"]["xlon0"]
                self.xlat0 = domain_dict["nam_conf_proj"]["xlat0"]
            else:
                raise KeyError("Missing keys3")
        else:
            raise KeyError("Missing key4")

        earth = 6.37122e6
        if self.xlat0 == 90.0 or self.xlat0 == -90.0:
            proj_string = (
                f"+proj=stere +lat_0={str(self.xlat0)} +lon_0={str(self.xlon0)} "
                f"+lat_ts={str(self.xlat0)}"
            )
        else:
            proj_string = (
                f"+proj=lcc +lat_0={str(self.xlat0)} +lon_0={str(self.xlon0)} "
                f"+lat_1={str(self.xlat0)} +lat_2={str(self.xlat0)} "
                f"+units=m +no_defs +R={str(earth)}"
            )

        logging.debug("Proj string: %s", proj_string)
        proj = pyproj.CRS.from_string(proj_string)
        wgs84 = pyproj.CRS.from_string("EPSG:4326")

        xloncen, xlatcen = pyproj.Transformer.from_crs(
            wgs84, proj, always_xy=True
        ).transform(self.xloncen, self.xlatcen)

        x_0 = float(xloncen) - (0.5 * ((float(self.nimax) - 1.0) * self.xdx))
        y_0 = float(xlatcen) - (0.5 * ((float(self.njmax) - 1.0) * self.xdy))
        self.x_0 = x_0
        self.y_0 = y_0
        xxx = np.empty([self.nimax])
        yyy = np.empty([self.njmax])
        # TODO vectorize
        for i in range(0, self.nimax):
            xxx[i] = x_0 + (float(i) * self.xdx)
        for j in range(0, self.njmax):
            yyy[j] = y_0 + (float(j) * self.xdy)
        self.xxx = xxx
        self.yyy = yyy
        y_v, x_v = np.meshgrid(yyy, xxx)
        logging.debug("x_v.shape=%s y_v.shape=%s", x_v.shape, y_v.shape)
        lons, lats = pyproj.Transformer.from_crs(proj, wgs84, always_xy=True).transform(
            x_v, y_v
        )

        logging.debug("lons.shape=%s lats.shape=%s", lons.shape, lats.shape)
        logging.debug("lons.shape=%s", lons)
        logging.debug("lats.shape=%s", lats)
        SurfexGeo.__init__(self, proj, lons, lats)

    def update_namelist(self, nml):
        """Update namelist.

        Args:
            nml (f90nml.Namelist): Namelist object.

        Returns:
            nml (f90nml.Namelist): Namelist object.
        """
        nml.update(
            {
                "nam_pgd_grid": {"cgrid": self.cgrid},
                "nam_conf_proj": {
                    "xlon0": self.xlon0,
                    "xlat0": self.xlat0,
                    "xrpk": math.sin(math.radians(self.xlat0)),
                    "xbeta": 0,
                },
                "nam_conf_proj_grid": {
                    "xlatcen": self.xlatcen,
                    "xloncen": self.xloncen,
                    "nimax": self.nimax,
                    "njmax": self.njmax,
                    "xdx": self.xdx,
                    "xdy": self.xdy,
                },
            }
        )
        if self.ilone is not None:
            nml["nam_conf_proj_grid"].update({"ilone": self.ilone})
        if self.ilate is not None:
            nml["nam_conf_proj_grid"].update({"ilate": self.ilate})
        if self.xtrunc is not None:
            nml["nam_conf_proj_grid"].update({"xtrunc": self.xtrunc})
        return nml

    def subset(self, geo):
        """Find subset of geo.

        Args:
            geo (surfex.Geo): Geometry to check.

        Returns:
            (list, list): lons, lats
        """
        lons = []
        lats = []
        if hasattr(geo, "cgrid") and geo.cgrid == self.cgrid:
            is_subset = True
            if self.xlon0 != geo.xlon0 or self.xlat0 != geo.xlat0:
                is_subset = False
            if self.xdx != geo.xdx:
                is_subset = False
            if self.xdy != geo.xdy:
                is_subset = False
            if self.nimax > geo.nimax:
                is_subset = False
            if self.njmax > geo.njmax:
                is_subset = False

            if is_subset:
                logging.info("Grids have same projection and grid spacing")
                x_0 = None
                y_0 = None

                for i in range(0, geo.nimax):
                    if round(self.x_0, 4) == round(geo.xxx[i], 4):
                        x_0 = i
                        break
                for j in range(0, geo.njmax):
                    if round(self.y_0, 4) == round(geo.yyy[j], 4):
                        y_0 = j
                        break
                if x_0 is not None and y_0 is not None:
                    logging.info(
                        "Grid is a subset of input grid %s %s", str(x_0), str(y_0)
                    )
                    lons = np.arange(x_0, x_0 + self.nimax, 1).tolist()
                    lats = np.arange(y_0, y_0 + self.njmax, 1).tolist()

        return lons, lats


class LonLatVal(SurfexGeo):
    """LonLatVal."""

    def __init__(self, from_json):
        """Construct a LonLatVal geometry.

        Used also for points/observations.

        Args:
            from_json (dict): Domain description,

        Raises:
            KeyError: Missing key
            KeyError: Missing keys

        """
        self.cgrid = "LONLATVAL"
        self.json = from_json
        domain_dict = BaseNamelist.lower_case_namelist_dict(from_json)

        if "nam_lonlatval" in domain_dict:
            if "xx" and "xy" and "xdx" and "xdy" in domain_dict["nam_lonlatval"]:
                self.x_x = domain_dict["nam_lonlatval"]["xx"]
                self.x_y = domain_dict["nam_lonlatval"]["xy"]
                self.xdx = domain_dict["nam_lonlatval"]["xdx"]
                self.xdy = domain_dict["nam_lonlatval"]["xdy"]
                proj4 = "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84"
                proj = pyproj.CRS.from_string(proj4)
                SurfexGeo.__init__(self, proj, np.asarray(self.x_x), np.asarray(self.x_y))
                self.can_interpolate = False
            else:
                raise KeyError("Missing keys")
        else:
            raise KeyError("Missing key")

    def update_namelist(self, nml):
        """Update namelist.

        Args:
            nml (f90nml.Namelist): Namelist object.

        Returns:
            nml (f90nml.Namelist): Namelist object.
        """
        nml.update(
            {
                "nam_pgd_grid": {"cgrid": self.cgrid},
                "nam_lonlatval": {
                    "xx": self.x_x,
                    "xy": self.x_y,
                    "xdx": self.xdx,
                    "xdy": self.xdy,
                },
            }
        )
        return nml

    def subset(self, geo):
        """Find subset of geo.

        Args:
            geo (surfex.Geo): Geometry to check.

        Returns:
            tuple: lons, lats
        """
        logging.info("Subset not implemented")
        lons = []
        lats = []
        return lons, lats


class Cartesian(SurfexGeo):
    """Cartesian."""

    def __init__(self, from_json):
        """Construct Cartesian geometry.

        Args:
            from_json (_type_): _description_

        Raises:
            KeyError: Missing key
            KeyError: Missing keys

        """
        self.cgrid = "CARTESIAN"
        self.json = from_json
        domain_dict = BaseNamelist.lower_case_namelist_dict(from_json)

        if "nam_cartesian" in domain_dict:
            if (
                "xlat0"
                and "xlon0"
                and "nimax"
                and "njmax"
                and "xdx"
                and "xdy" in domain_dict["nam_cartesian"]
            ):
                self.xlat0 = domain_dict["nam_cartesian"]["xlat0"]
                self.xlon0 = domain_dict["nam_cartesian"]["xlon0"]
                self.nimax = domain_dict["nam_cartesian"]["nimax"]
                self.njmax = domain_dict["nam_cartesian"]["njmax"]
                self.xdx = domain_dict["nam_cartesian"]["xdx"]
                self.xdy = domain_dict["nam_cartesian"]["xdy"]
                proj = None
                lons = []
                lats = []
                for i in range(0, self.nimax):
                    lons.append(self.xlon0 + i * self.xdx)
                for j in range(0, self.njmax):
                    lats.append(self.xlat0 + j * self.xdy)

                SurfexGeo.__init__(self, proj, np.asarray(lons), np.asarray(lats))
            else:
                raise KeyError("Missing keys")
        else:
            raise KeyError("Missing key")

    def update_namelist(self, nml):
        """Update namelist.

        Args:
            nml (f90nml.Namelist): Namelist object.

        Returns:
            nml (f90nml.Namelist): Namelist object.
        """
        print(nml)
        nml.update(
            {
                "nam_pgd_grid": {"cgrid": self.cgrid},
                "nam_cartesian": {
                    "xlat0": self.xlat0,
                    "xlon0": self.xlon0,
                    "nimax": self.nimax,
                    "njmax": self.njmax,
                    "xdx": self.xdx,
                    "xdy": self.xdy,
                },
            }
        )
        return nml

    def subset(self, geo):
        """Find subset of geo.

        Args:
            geo (surfex.Geo): Geometry to check.

        Returns:
            tuple: lons, lats

        """
        logging.info("Subset not implemented")
        lons = []
        lats = []
        return lons, lats


class LonLatReg(SurfexGeo):
    """LonLatReg."""

    def __init__(self, from_json):
        """Construct the LonLatReg geometry.

        Args:
            from_json (dict): Domain definition.

        Raises:
            KeyError: Missing key
            KeyError: Missing keys
            ZeroDivisionError: nlon and/or nlat is 0

        """
        self.cgrid = "LONLAT REG"
        self.json = from_json
        domain_dict = BaseNamelist.lower_case_namelist_dict(from_json)

        if "nam_lonlat_reg" in domain_dict:
            if (
                "xlonmin"
                and "xlonmax"
                and "xlatmin"
                and "xlatmax"
                and "nlon"
                and "nlat" in domain_dict["nam_lonlat_reg"]
            ):
                self.xlonmin = domain_dict["nam_lonlat_reg"]["xlonmin"]
                self.xlonmax = domain_dict["nam_lonlat_reg"]["xlonmax"]
                self.xlatmin = domain_dict["nam_lonlat_reg"]["xlatmin"]
                self.xlatmax = domain_dict["nam_lonlat_reg"]["xlatmax"]
                self.nlon = domain_dict["nam_lonlat_reg"]["nlon"]
                self.nlat = domain_dict["nam_lonlat_reg"]["nlat"]
            else:
                raise KeyError("Missing keys")
        else:
            raise KeyError("Missing key")

        proj_string = "+proj=longlat +datum=WGS84 +no_defs +ellps=WGS84"
        proj = pyproj.CRS.from_string(proj_string)
        lons = []
        lats = []
        if self.nlon == 0 or self.nlat == 0:
            raise ZeroDivisionError("nlon and/or nlat is 0")

        dlon = (self.xlonmax - self.xlonmin) / (self.nlon - 1)
        dlat = (self.xlatmax - self.xlatmin) / (self.nlat - 1)
        logging.debug("nlon=%s nlat=%s dlon=%s dlat=%s", self.nlon, self.nlat, dlon, dlat)
        for i in range(0, self.nlon):
            lons.append(self.xlonmin + i * dlon)
        for j in range(0, self.nlat):
            lats.append(self.xlatmin + j * dlat)

        # proj, npoints, nlons, nlats, lons, lats
        latitudes, longitudes = np.meshgrid(lats, lons)
        logging.debug("longitudes=%s latitudes=%s", longitudes.shape, latitudes.shape)
        SurfexGeo.__init__(self, proj, longitudes, latitudes)

    def update_namelist(self, nml):
        """Update namelist.

        Args:
            nml (f90nml.Namelist): Namelist object.

        Returns:
            nml (f90nml.Namelist): Namelist object.
        """
        nml.update(
            {
                "nam_pgd_grid": {"cgrid": self.cgrid},
                "nam_lonlat_reg": {
                    "xlonmin": self.xlonmin,
                    "xlonmax": self.xlonmax,
                    "xlatmin": self.xlatmin,
                    "xlatmax": self.xlatmax,
                    "nlon": self.nlon,
                    "nlat": self.nlat,
                },
            }
        )
        return nml

    def subset(self, geo):
        """Find subset of geo.

        Args:
            geo (surfex.Geo): Geometry to check.

        Returns:
            tuple: lons, lats

        """
        logging.info("Subset not implemented")
        lons = []
        lats = []
        return lons, lats


class IGN(SurfexGeo):
    """IGN."""

    def __init__(self, from_json, recreate=False):
        """Construct a IGN geometry.

        Args:
            from_json (dict): Domain definition.
            recreate (bool, optional): Recreate the cached mask. Defaults to False.

        Raises:
            NotImplementedError: Projection not implemented
            KeyError: Missing key
            KeyError: Missing keys

        """
        self.cgrid = "IGN"
        self.json = from_json
        domain_dict = BaseNamelist.lower_case_namelist_dict(from_json)

        if "nam_ign" in domain_dict:
            if (
                "clambert"
                and "npoints"
                and "xx"
                and "xy"
                and "xdx"
                and "xdy"
                and "xx_llcorner"
                and "xy_llcorner"
                and "xcellsize"
                and "ncols"
                and "nrows" in domain_dict["nam_ign"]
            ):

                self.clambert = domain_dict["nam_ign"]["clambert"]
                npoints = domain_dict["nam_ign"]["npoints"]
                self.x_x = domain_dict["nam_ign"]["xx"]
                self.x_y = domain_dict["nam_ign"]["xy"]
                self.xdx = domain_dict["nam_ign"]["xdx"]
                self.xdy = domain_dict["nam_ign"]["xdy"]
                self.xx_llcorner = domain_dict["nam_ign"]["xx_llcorner"]
                self.xy_llcorner = domain_dict["nam_ign"]["xy_llcorner"]
                self.xcellsize = domain_dict["nam_ign"]["xcellsize"]
                self.ncols = domain_dict["nam_ign"]["ncols"]
                self.nrows = domain_dict["nam_ign"]["nrows"]
            else:
                raise KeyError("Missing keys")
        else:
            raise KeyError("Missing key")

        if self.clambert == 7:
            proj4 = (
                "+proj=lcc +lat_0=63.5 +lon_0=15.0 +lat_1=63.5 +lat_2=63.5 "
                "+no_defs +R=6.37122e+6"
            )
            self.xloncen = 17
            self.xlatcen = 63.0
            self.xlon0 = 15
            self.xlat0 = 63.5
        else:
            raise NotImplementedError

        proj = pyproj.CRS.from_string(proj4)
        wgs84 = pyproj.CRS.from_string("EPSG:4326")

        pxall = self.get_coord(self.x_x, self.xdx, "x", recreate)
        pyall = self.get_coord(self.x_y, self.xdy, "y", recreate)
        self.mask = self.ign_mask(pxall, pyall, self.x_x, self.x_y, recreate)

        # proj, npoints, nlons, nlats, lons, lats
        lons = []
        lats = []
        for i in range(0, npoints):
            lon, lat = pyproj.Transformer.from_crs(proj, wgs84, always_xy=True).transform(
                self.x_x[i], self.x_y[i]
            )
            lons.append(lon)
            lats.append(lat)

        SurfexGeo.__init__(self, proj, np.asarray(lons), np.asarray(lats))

    @staticmethod
    def get_coord(pin, pdin, coord, recreate=False):
        """Get the IGN coordinates.

        Args:
            pin (list): _description_
            pdin (list): _description_
            coord (list): _description_
            recreate (bool, optional): _description_. Defaults to False.

        Returns:
            list: Output coordinates

        """
        pout = []
        cache = "/tmp/." + coord + "_cached"  # noqa S108
        if os.path.isfile(cache) and not recreate:
            with open(cache, mode="r", encoding="utf-8") as file_handler:
                cached_coord = file_handler.read().splitlines()
            for c_coord in cached_coord:
                pout.append(float(c_coord))
            return pout

        zdout = []
        ksize = 0
        if len(pin) > 0:
            zdout.append(float(pdin[0]) / 2.0)
            pout.append(pin[0])
            ksize = 1
            if len(pin) > 1:
                ksize = 2
                pout.append(pin[0] - pdin[0])
                zdout.append(0.0)
            if len(pin) > 2:
                ksize = 3
                pout.append(pin[0] + pdin[0])
                zdout.append(0.0)

        for i, pinval in enumerate(pin):
            for j in range(0, ksize):
                if pout[j] == pinval:
                    break
                if j == ksize - 1:
                    ksize = ksize + 1
                    pout.append(pinval)
                    zdout.append(float(pdin[i]) / 2.0)

            # Mesh constrains
            for j in range(0, ksize):
                if pout[j] < pin[i] and (pout[j] + zdout[j]) >= (pin[i] - pdin[i]):
                    break
                if j == ksize - 1:
                    ksize = ksize + 1
                    pout.append(pin[i] - pdin[i])
                    zdout.append(0.0)

            for j in range(0, ksize):
                if pout[j] > pin[i] and (pout[j] - zdout[j]) <= (pin[i] + pdin[i]):
                    break
                if j == ksize - 1:
                    ksize = ksize + 1
                    pout.append(pin[i] + pdin[i])
                    zdout.append(0.0)

        # Sort pout
        pout = sorted(pout)

        with open(cache, mode="w", encoding="utf-8") as file_handler:
            for pout_val in pout:
                file_handler.write(str(pout_val) + "\n")
        logging.info("Cached coordinates for : %s", coord)
        return pout

    @staticmethod
    def ign_mask(pxall, pyall, xxx, yyy, recreate):
        """Create the IGN mask.

        Args:
            pxall (_type_): _description_
            pyall (_type_): _description_
            xxx (_type_): _description_
            yyy (_type_): _description_
            recreate (_type_): _description_

        Raises:
            Exception: _description_

        Returns:
            _type_: _description_
        """
        mask = []

        cache = "/tmp/.mask"  # noqa S108
        if os.path.isfile(cache) and not recreate:
            with open(cache, mode="r", encoding="utf-8") as file_handler:
                cached_mask = file_handler.read().splitlines()

            for cached_mask_ind in cached_mask:
                mask.append(int(cached_mask_ind))

            if len(mask) != len(xxx) or len(mask) != len(yyy):
                raise Exception("Cached mask mismatch! ", len(mask), len(xxx), len(yyy))

            return mask

        logging.warning("Creating mask. This takes time:")
        count = -1

        for i, pxall_val in enumerate(pxall):
            for pyall_val in pyall:

                count = count + 1
                for k, xval in enumerate(xxx):
                    if xval == pxall_val and yyy[k] == pyall_val:
                        mask.append(count)
                        break

            logging.debug("%s/%s", i, len(pxall))

        # Cache mask for later use
        with open(cache, mode="w", encoding="utf-8") as file_handler:
            for mask_ind in mask:
                file_handler.write(str(mask_ind) + "\n")

        logging.info("Created mask: %s", mask)
        return mask

    def update_namelist(self, nml):
        """Update namelist.

        Args:
            nml (f90nml.Namelist): Namelist object.

        Returns:
            nml (f90nml.Namelist): Namelist object.
        """
        nml.update(
            {
                "nam_pgd_grid": {"cgrid": self.cgrid},
                "nam_ign": {
                    "clambert": self.clambert,
                    "npoints": self.npoints,
                    "xx": self.x_x,
                    "xy": self.x_y,
                    "xdx": self.xdx,
                    "xdy": self.xdy,
                    "xx_llcorner": self.xx_llcorner,
                    "xy_llcorner": self.xy_llcorner,
                    "xcellsize": self.xcellsize,
                    "ncols": self.ncols,
                    "nrows": self.nrows,
                },
            }
        )
        return nml

    def subset(self, geo):
        """Find subset of geo.

        Args:
            geo (surfex.Geo): Geometry to check.

        Returns:
            tuple: lons, lats

        """
        logging.info("Subset not implemented")
        lons = []
        lats = []
        return lons, lats


def get_geo_object(from_json):
    """Get a surfex geometry object from a dictionary.

    Args:
        from_json (dict): Domain definition.

    Raises:
        NotImplementedError: Grid not implemented
        KeyError: Missing grid information cgrid
        KeyError: nam_pgd_grid not set!

    Returns:
        surfex.Geo: Surfex geometry.

    """
    domain_dict = {}
    for key in from_json:
        lower_case_dict = {}
        for key2 in from_json[key]:
            lower_case_dict.update({key2.lower(): from_json[key][key2]})
        domain_dict.update({key.lower(): lower_case_dict})

    if "nam_pgd_grid" in domain_dict:
        if "cgrid" in domain_dict["nam_pgd_grid"]:
            cgrid = domain_dict["nam_pgd_grid"]["cgrid"]
            if cgrid == "CONF PROJ":
                return ConfProj(from_json)
            if cgrid == "LONLATVAL":
                return LonLatVal(from_json)
            if cgrid == "LONLAT REG":
                return LonLatReg(from_json)
            if cgrid == "IGN":
                return IGN(from_json)
            if cgrid == "CARTESIAN":
                return Cartesian(from_json)
            raise NotImplementedError(f"CGRID={cgrid} is not implemented")

        raise KeyError("Missing grid information cgrid")
    raise KeyError("nam_pgd_grid not set!")


def set_domain(settings, domain, hm_mode=False):
    """Set domain.

    Args:
        settings (dict): Domain definitions
        domain (str): Domain name
        hm_mode (bool, optional): Harmonie definition. Defaults to False.

    Raises:
        KeyError: Domain not found
        ValueError: Settings should be a dict

    Returns:
        dict: Domain dictionary

    """
    if isinstance(settings, dict):
        if domain in settings:
            if hm_mode:

                ezone = 11
                if "EZONE" in settings[domain]:
                    ezone = settings[domain]["EZONE"]

                domain_dict = {
                    "nam_pgd_grid": {"cgrid": "CONF PROJ"},
                    "nam_conf_proj": {
                        "xlat0": settings[domain]["LAT0"],
                        "xlon0": settings[domain]["LON0"],
                    },
                    "nam_conf_proj_grid": {
                        "ilone": ezone,
                        "ilate": ezone,
                        "xlatcen": settings[domain]["LATC"],
                        "xloncen": settings[domain]["LONC"],
                        "nimax": settings[domain]["NLON"] - ezone,
                        "njmax": settings[domain]["NLAT"] - ezone,
                        "xdx": settings[domain]["GSIZE"],
                        "xdy": settings[domain]["GSIZE"],
                    },
                }
            else:
                domain_dict = settings[domain]
            return domain_dict

        logging.error("Domain not found: %s", domain)
        raise KeyError("Domain not found: " + domain)
    logging.error("Settings should be a dict")
    raise ValueError("Settings should be a dict")


def shape2ign(catchment, infile, output, ref_proj, indent=None):
    """Read a shape file and convert to IGN geo.

    Args:
        catchment (_type_): _description_
        infile (_type_): _description_
        output (_type_): _description_
        ref_proj (_type_): _description_
        indent (_type_, optional): _description_. Defaults to None.

    """
    from_json = json.load(open(ref_proj, mode="r", encoding="utf-8"))
    geo = get_geo_object(from_json)
    earth = 6.37122e6
    proj_string = (
        f"+proj=lcc +lat_0={str(geo.xlat0)} +lon_0={str(geo.xlon0)} "
        f"+lat_1={str(geo.xlat0)} +lat_2={str(geo.xlat0)} "
        f"+units=m +no_defs +R={str(earth)}"
    )

    logging.debug(proj_string)
    proj = pyproj.CRS.from_string(proj_string)
    wgs84 = pyproj.CRS.from_string("EPSG:4326")

    shpfile = ogr.Open(infile)
    shape = shpfile.GetLayer(0)

    # TODO find index
    logging.info("TODO: find %s index", catchment)
    feature = shape.GetFeature(2562)
    feature_dict = json.loads(feature.ExportToJson())
    logging.debug(feature_dict["properties"]["stNavn"])
    logging.debug(feature_dict)

    lons = []
    lats = []
    values = []
    for point in feature_dict["geometry"]["coordinates"][0]:
        lons.append(point[0])
        lats.append(point[1])
        values.append(point[2])

    xxx, yyy = pyproj.Transformer.from_crs(wgs84, proj, always_xy=True).transform(
        lons, lats
    )
    x_1 = min(xxx)
    x_2 = max(xxx)
    y_1 = min(yyy)
    y_2 = max(yyy)
    logging.debug("%s %s %s %s", x_1, x_2, y_1, y_2)
    ring = ogr.Geometry(ogr.wkbLinearRing)
    for point, xval in enumerate(xxx):
        ring.AddPoint(xval, yyy[point])

    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)

    ign_x = []
    ign_y = []
    delta_x = 1000
    delta_y = 1000
    n_x = int((x_2 - x_1) / delta_x) + 1
    n_y = int((y_2 - y_1) / delta_y) + 1
    xdx = []
    xdy = []

    npoints = 0
    for x_p in range(0, n_x):
        xxx = x_1 + (x_p * delta_x) - delta_x
        for y_p in range(0, n_y):
            yyy = y_1 + (y_p * delta_y) - delta_y
            point = ogr.Geometry(ogr.wkbPoint)
            point.AddPoint(xxx, yyy)
            if not poly.Intersection(point).IsEmpty():
                ign_x.append(xxx)
                xdx.append(delta_x)
                ign_y.append(yyy)
                xdy.append(delta_y)
                npoints = npoints + 1

    nam_json = {
        "nam_pgd_grid": {"cgrid": "IGN"},
        "nam_ign": {
            "clambert": 7,
            "npoints": npoints,
            "xx": ign_x,
            "xy": ign_y,
            "xdx": xdx,
            "xdy": xdy,
            "xx_llcorner": 0,
            "xy_llcorner": 0,
            "xcellsize": "250",
            "ncols": 0,
            "nrows": 0,
        },
    }
    with open(output, "w", encoding="utf-8") as file_handler:
        json.dump(nam_json, file_handler, indent=indent)
